#include <xc.h>
#include "init.h"
#include "main.h"
#include "spi.h"
#include "uart.h"

extern unsigned char deviceIsKeyfob;

void Initial(void)
{
    // Oscillator configuration
    //Using internal oscillator with FRCDIV (8Mhz oscillator divided to 4 MHz)
    // Fosc = 4Mhz
    // Fcy = 2Mhz 

    // I/O Port configuration
    PORTA = 0b0000000000000000;  //Set initial value for PORTA
    TRISA = 0b1111111111101100;  //Set I/O for PORTA as input (1) or output (0)

    PORTB = 0b0000000000000000;  //Set initial value for PORTB
    TRISB = 0b0100111111111111;  //Set I/O for PORTB as input (1) or output (0)

    // disable analog inputs (use digital)
    ANSA = 0;
    ANSB = 0; 

    if (deviceIsKeyfob) {
        // Internal Pull-up config
        // BUTTON0 = CN30, B1 = CN22, B2 = CN1, B3 = CN29, B4 = CN23
        CNPU1 = 0b0000000000000010;
        CNPU2 = 0b0110000011000000;

        // Input Change Notifications
        CNEN1 = 0b0000000000000010;
        CNEN2 = 0b0110000011000000;
        
        PWR_ON_TRIS = 0;    // enable line for switching regulator
        PWR_ON = 0;
    } else {
        GPIO1_TRIS = 0;   // GPIO0
        GPIO2_TRIS = 0;
        TRISBbits.TRISB7 = 0;   // RB7 is an output for U1TX
         
        // UART configuration
        U1BRG = 12;  // 9600 baud @ Fcy = 2MHz: U1BRG = 2e6/(16*9600)-1 = 12

        U1MODEbits.UARTEN = 1;      // enable UART
        U1STAbits.UTXISEL0 = 1;     // TX interrupt selection
        U1STAbits.UTXISEL1 = 0;
        IPC3bits.U1TXIP = 6;
        IEC0bits.U1TXIE = 1;        // enable TX interrupt
        U1STAbits.UTXEN = 1;        // enable transmit
        IFS0bits.U1TXIF = 0;        // clear TX interupt flag

        U1STAbits.URXISEL = 0b00;   // RX interrupt on any character
        IEC0bits.U1RXIE = 1;        // enable RX interrupt
        IFS0bits.U1RXIF = 0;        // clear RX interupt flag
        IPC2bits.U1RXIP = 0b111;
    }
    CNPU2bits.CN30PUE = 1;  // GPIO 0 pullup
    
    IFS1bits.CNIF = 0;  // clear input change interrupt flag
    IEC1bits.CNIE = 1;  // enable input change interrupt
    IPC4bits.CNIP = 7;  // set interrupt priority to max

    // SPI config
    SPI1CON1 = 0b0000000100111111;  // SPI master mode, 1:1 prescalar
    SPI1STATbits.SPIEN = 1; // enable SPI
    SPI1STATbits.SPIROV = 0;   // clear SPI errors
    SPI_CS = 1;     // CS pin must be normally high


    // ADC init
    AD1CON1bits.SSRC = 0b0111;       // clearing SAMP bit ends sampling and beings conversion
    AD1CON1bits.ASAM = 1;       // sampling immediately after conversion

    AD1CON2bits.PVCFG = 0b00;   // use AVdd as ref
    AD1CON2bits.SMPI = 0;       // interrupt at each conversion

    AD1CON3bits.ADCS = 0;
    AD1CON3bits.SAMC = 31;//0b0100;  // auto sample time = 4 * TAD

    AD1CON5bits.BGREQ = 1;      // band gap enabled
    AD1CHSbits.CH0SA = 0b11010; // input to ADC is internal band gap reference
    //IFS0bits.AD1IF = 0;
    //IEC0bits.AD1IE = 1;
}

void initTimers()
{
    // TIMER 1 PERIOD (with Fosc=4Mhz, 1:8 prescale) = 0.262s
    IEC0bits.T1IE = 1;          // timer 1 interrupt enable
    IPC0bits.T1IP = 6;      // timer 1 interrupt priority
    T1CONbits.TCKPS = 1;     // clock prescale is 1:8
    T1CONbits.TON = 1;          // enable timer 1

    // TIMER 2 PERIOD (with Fosc=4Mhz, 1:1 prescale) = 32.768 ms
    IEC0bits.T2IE = 1;          // timer 2 interrupt enable
    IPC1bits.T2IP = 5;      // timer 2 interrupt priority 

    // TIMER 3 PERIOD (with Fosc=4Mhz, 1:64 prescale) = ~2 s
    IEC0bits.T3IE = 1;          // timer 3 interrupt enable
    IPC2bits.T3IP = 5;      // timer 3 interrupt priority 
    T3CONbits.TCKPS = 2;

    // TIMER 4 (with Fosc=4Mhz, 1:64 prescale) 
    IEC1bits.T4IE = 1;          // timer 4 interrupt enable
    IPC6bits.T4IP = 4;          // timer 4 interrupt priority
    T4CONbits.TCKPS = 2;
    T4CONbits.TON = 1;
    //PR4 = 15625;    // interrupt every .5s

    // TIMER 5 PERIOD (with Fosc=4Mhz, 1:1 prescale) = 32.768 ms
    IEC1bits.T5IE = 1;          // timer 2 interrupt enable
    IPC7bits.T5IP = 5;      // timer 2 interrupt priority
    PR5 = 2000;     // interrupt every 1ms
}