// Nali Guardian Firmware
// Daniel Smith
// www.pagemac.com

#include <xc.h>
#include <stdio.h>
#include "AES.h"
#include "config_bits.h"
#include "init.h"
#include "spi.h"
#include "main.h"
#include "rf.h"
#include "nalig_protocol.h"
#include "sys_functions.h"
#include "uart.h"
volatile int buttonFlag=0;
volatile unsigned char nextLED = OFF, blinkLED = OFF, currentLED;
volatile unsigned char numBlinks = 0;

volatile unsigned char IRQ_flag=0;
volatile unsigned char timer1_flag=0;
volatile unsigned char timer2_flag=0; 
volatile unsigned char timer3_flag=0;
volatile unsigned char timer4_flag=0;
volatile unsigned char sleep_flag=0;
volatile unsigned char timer5_flag=1;

extern unsigned char cacheHasEntries;

//volatile unsigned char motionX_counts=0;
//volatile unsigned char motionX_flag=0;
//unsigned char last_motionX_counts;

unsigned char buttonLastState[5] = {1, 1, 1, 1, 1};
int buttonPressed=BUTTON_UNPRESSED;

unsigned char buttonCode[5] = {0, 0, 0, 0};
unsigned char buttonsPressed=0;
volatile unsigned char menuLevel=0;   
unsigned char lastDataRequestor=ADDR_MASTER;  // node address of last data requestor

//unsigned char mfg_id[6] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

unsigned char challenge[16];
volatile unsigned char sensorACK=0;

extern unsigned char currentPowerLevel;

unsigned char lastButton1=1;
unsigned char ledToggle;

volatile unsigned char UART_RX_buffer[UART_BUFFER_SIZE];
volatile unsigned char UART_RX_buffer_size;

extern volatile unsigned char tx_buffer[TX_BUFFER_SIZE];
extern volatile unsigned char tx_buffer_pos, tx_buffer_len, tx_fifo;
extern unsigned char securityState;

unsigned char assignType=0, assignAddr=0;

extern unsigned char handshake_cmd[2];
extern unsigned char receivedPackets;
unsigned int tempcount;

extern unsigned char rnd_base[16];
extern unsigned char rnd_counter;
volatile unsigned int rnd_adder;

unsigned char sensorDelay = 0;
unsigned char doorSensorLast[3];
unsigned char bSensorsEnabled = 1;
unsigned char ignore_non_relayed_packets=0;

sys_config myConfig;
unsigned char unsavedConfigChanges=0;
unsigned char deviceIsKeyfob;
unsigned char badAddrFlag;


unsigned char tempFlag;
nalig_msg tx_msg;

int main()
{
    unsigned int ret;
    //while (1) {}
    initTimers();

    if (!readDeviceConfig())    // did we read a valid config from EEPROM?
    {
        loadDeafultConfig();    // load defaults
        saveDeviceConfig();     // save defaults
    }
    
    /*
    myConfig.myAddr = 0x02;
    myConfig.myType = 0x02;
    saveDeviceConfig();
    //*/

    //myConfig.hasRFE = 0;
    Initial();
    
    setLED(OFF);

    CYRF6936_init();
    CYRF6936_sleep();
    
    if (CYRF6936_mfgid())   // read the MFG ID from the CYRF6936 t
    {
        setLED2(WHITE, GREEN);      // valid!
    } else {    // not a valid MFG ID.  hardware is not connected properly. we can't do anything.
        blinkLED(RED, 4);
        while (1)
        { Idle(); } // sleep forever, sweet prince
    }

    if (myConfig.myAddr == ADDR_BROADCAST || myConfig.myAddr == ADDR_TEMP ||
        myConfig.myAddr == 0x00 || myConfig.myType == 0xFF)    // must be a new node, we don't have a valid address
    {
        myConfig.myAddr = ADDR_TEMP;
        myConfig.myType = ADDR_TEMP;
        badAddrFlag = 1;
        blinkLED(WHITE, 3);
    }


    if (!deviceIsKeyfob)    // non-keyfobs should do this stuff
    {
        setGPIO(myConfig.GPIO_config1, myConfig.GPIO_config2);
        T3CONbits.TON = 1;          // enable timer 3
        //GPIO0_TRIS = 0;
        
        txStr("\rNali Guardian System v");
        txHex(FIRMWARE_VER_MAJ);
        txVal(".", FIRMWARE_VER_MIN);
        generateNonce(challenge);

       CYRF6936_wakeup();

        if (myConfig.myAddr != ADDR_MASTER) {
            makePacket(&tx_msg, ADDR_MASTER, FLAGS_NONE, CMD_STATUS, STATUS_ONLINE);
            if (sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD)) {
                setLED2(WHITE, GREEN);      // master heard us
            } else {
                setLED2(WHITE, BLUE);      // master didn't hear us
            }
        }

        for (ret=0; ret<16; ret++)
            txHex(challenge[ret]);
        txChar('\r');

        txReport(myConfig.myAddr, ":a", myConfig.myType);   // print current address
    }

    initCache();

    if (myConfig.myType == TYPE_DOOR)
    {
        doorSensorLast[0] = GPIO1;
        doorSensorLast[1] = GPIO3;
    }
    /*else if (myConfig.myType == TYPE_MOTIONX)
    {
        CNEN2bits.CN29IE = 1;   // GPIO3 change interrupt
        T4CONbits.TON = 1;      // enable timer 4
    }*/
    //tempFlag = 1;
    //buttonFlag = 1;
    UART_RX_buffer_size =0;
    while (1)
    {


        if (deviceIsKeyfob) {
            keyfobMainColors();
        } else {
            nodeMainLoop();
        }

        /*if (sleep_flag)
        {
            sleep_flag = 0;
            Sleep();
        }*/

        if (timer1_flag)    // timer 1 used for LED
        {
            timer1_flag = 0;

            setLED(nextLED);        // set the LED to the next state

            if (nextLED != OFF)     // next state was some other color
            {
                nextLED = OFF;  // after this color, we will turn off
                TMR1 = 0x4000;  // make this one shorter
                Idle();
                Nop();
            }
            else    // the next LED state was OFF
            {
                if (numBlinks > 0)  // are we blinking the LED?
                {
                    numBlinks--;
                    nextLED = blinkLED; // ok after this, we are turning it back on
                    TMR1 = 0x4000;
                }
                else    // nope, just done doing stuff
                {
                    //if (deviceIsKeyfob && (buttonsPressed == 0)) {  // for keyfobMainLoopAdv
                    if (deviceIsKeyfob && (menuLevel == 0)) {  // for keyfobMainLoopColors
                        Sleep();
                        Nop();
                    }
                    T1CONbits.TON = 0;  // turn off timer 1

                }
            }
        }

        if (timer4_flag)
        {
            if (cacheHasEntries) initCache();   // flush the cache after 2s of idle
            timer4_flag = 0;
        }
        
        //ClrWdt();
    }

   return 0;
}

void keyfobMainColors()
{
   unsigned char returnAddr;
   unsigned char num=0;

   unsigned char button;

    if (buttonFlag)
   {
       // use timer 2 to delay for button debouncing
        T1CONbits.TON = 0;  // disable timer 1 so we don't sleep forever during debounce
        IEC1bits.CNIE = 0;  // disable input change interrupt
        T2CONbits.TON = 1;          // enable timer 2
        TMR2 = 0;
        Idle(); Nop();
        T2CONbits.TON = 0;          // disable timer 2
        IEC1bits.CNIE = 1;  // enable input change interrupt
        T1CONbits.TON = 1;

        // check which buttons are pressed down
       if (BUTTON0 == 0 && buttonLastState[0])                // down
       {
            button = 3;
       } else if (BUTTON1 == 0 && buttonLastState[1]) {       // up
            button = 1;
       } else if (BUTTON2 == 0 && buttonLastState[2]) {       // right
            button = 2;
       } else if (BUTTON3 == 0 && buttonLastState[3]) {       // left
           button = 4;
       } else if (BUTTON4 == 0 && buttonLastState[4]) {       // center
            button = 5;
       } else {
           button = 0;  // probably just a button being released
       }

        buttonFlag = 0;

        buttonLastState[0] = BUTTON0;
        buttonLastState[1] = BUTTON1;
        buttonLastState[2] = BUTTON2;
        buttonLastState[3] = BUTTON3;
        buttonLastState[4] = BUTTON4;

        if (!button)  goto finish;

        

        switch (menuLevel)
        {
            case 0:     // root menu
                menuLevel = button; 

                if (button == 1) {        setLEDfast(PURPLE);
                } else if (button == 2) {  setLEDfast(BLUE);
                } else if (button == 3) {  setLEDfast(TEAL);
                } else if (button == 4) {  setLEDfast(GREEN);
                } else if (button == 5) {  setLEDfast(YELLOW); }
                break;
                
            case 1:     // sounds
                if (button == 1) {         num=38;
                } else if (button == 2) {  num=100;
                } else if (button == 3) {  num=63;
                } else if (button == 4) {  num=53;
                } else if (button == 5) {  num=70; }

                CYRF6936_wakeup();
                makePacket(&tx_msg, TYPE_AUDIO, FLAG_BROADCAST + FLAG_RELAY, CMD_PLAY_AUDIO, num);
                if (sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD)) {
                    setLED(BLUE);
                } else {
                    setLED(YELLOW);
                }
                CYRF6936_sleep();
                menuLevel = 0;
                break;
                
            case 2:     // automation

                if (button == 1) {
                    CYRF6936_wakeup();
                    makePacket(&tx_msg, 0x60, FLAGS_NONE, CMD_PULSE_GPIO, 0x02);
                    if (sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD)) {
                        setLED(BLUE);
                    } else {
                        setLED(YELLOW);
                    }
                    CYRF6936_sleep();
                }
                else if (button == 2)
                {
                    num = 1;
                    goto open_garage_door;
                }
                else if (button == 4)
                {
                    num = 2;
                    open_garage_door:
                    CYRF6936_wakeup();
                    makePacket(&tx_msg, 0xC0, FLAGS_NONE, CMD_PULSE_GPIO, num);
                    if(sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD)) {
                        setLED(GREEN);
                    } else {
                        setLED(YELLOW);
                    }
                    CYRF6936_sleep();
                }
                else {
                 setLED(RED);
                }


                menuLevel = 0;   
                break;
            case 4: // security commands

                if (button == 1) {         num=0x03;    // arm system
                } else if (button == 2) {  num=0x10;    // toggle security state
                } else if (button == 3) {  num=0x00;    // disarm system
                } else if (button == 4) {  num=0xF0; goto check_status;    // check state
                } else if (button == 5) {  num=0xF0; goto check_status; }
                
                setLEDfast(BLUE);

                // execute a handshake for arm, toggle, or disarm system (depending on the button)
                CYRF6936_wakeup();
                makePacket(&tx_msg, ADDR_MASTER, FLAGS_NONE, CMD_SET_SEC_STATE, num);
                handshake_cmd[0] = CMD_SET_SEC_STATE;
                returnAddr = sendNaligPacket(FREQ_A, &tx_msg, PAYLOAD_LEN_SM);  // send security toggle request
                if (returnAddr) {
                    if (receiveNaligPacket(FREQ_A, PROCESS_PACKETS))  {// receive and process challenge
                        if (!receiveNaligPacket(FREQ_A, PROCESS_PACKETS)) {   // receive and process access denied/granted message
                            setLED2(YELLOW, TEAL);
                        }
                    } else {
                        setLED2(YELLOW, BLUE);
                    }
                } else {
                    setLED(YELLOW);
                }
                CYRF6936_sleep();
                menuLevel = 0;
                break;

                check_status:
                CYRF6936_wakeup();
                makePacket(&tx_msg, ADDR_MASTER, FLAGS_NONE, CMD_SET_SEC_STATE, num);
                returnAddr = sendNaligPacket(FREQ_A, &tx_msg, PAYLOAD_LEN_SM);  // send security toggle request
                if (returnAddr) {
                    if (receiveNaligPacket(FREQ_A, PROCESS_PACKETS))  {// receive and process response
                        // led will be controlled from there
                    } else {
                        setLED2(YELLOW, BLUE);
                    }
                } else {
                    setLED(YELLOW);
                }
                CYRF6936_sleep();
                menuLevel = 0;
                break;

            case 5: // remote command
                if (button == 1) {         num=0x50;
                } else if (button == 2) {  num=0x51;
                } else if (button == 3) {  num=0x52;
                } else if (button == 4) {  num=0x53;
                } else if (button == 5) {  num=0x5A; }

                CYRF6936_wakeup();
                makePacket(&tx_msg, ADDR_MASTER, FLAG_RELAY, CMD_SPECIAL, num);
                if (sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD)) {
                    setLED(BLUE);
                } else {
                    setLED(YELLOW);
                }
                CYRF6936_sleep();
                menuLevel = 0;
                break;

            default:
                setLED(RED);
                menuLevel = 0;
        }



        TMR3 = 0;
        T3CONbits.TON = 1;

        finish:
        Idle();
    }
}

void keyfobMainLoopAdv()
{
    /*
   //int ret;
   unsigned char returnAddr;
   unsigned long code;
   unsigned char buttonUp=0;
   unsigned char num;

   if (buttonFlag)
   {
       // use timer 2 to delay for button debouncing
        T1CONbits.TON = 0;  // disable timer 1 so we don't sleep forever during debounce
        IEC1bits.CNIE = 0;  // disable input change interrupt
        T2CONbits.TON = 1;          // enable timer 2
        TMR2 = 0;
        Idle(); Nop();
        T2CONbits.TON = 0;          // disable timer 2
        IEC1bits.CNIE = 1;  // enable input change interrupt
        T1CONbits.TON = 1;
        
        // check which buttons are pressed down
       if (BUTTON0 == 0 && buttonLastState[0])                // down
       {
            buttonCode[buttonsPressed++] = 3;
       } else if (BUTTON1 == 0 && buttonLastState[1]) {       // up
            buttonCode[buttonsPressed++] = 1;
       } else if (BUTTON2 == 0 && buttonLastState[2]) {       // right
            buttonCode[buttonsPressed++] = 2;
       } else if (BUTTON3 == 0 && buttonLastState[3]) {       // left
            buttonCode[buttonsPressed++] = 4;
       } else if (BUTTON4 == 0 && buttonLastState[4]) {       // center
            buttonCode[buttonsPressed++] = 5;
       } else {
           buttonUp = 1;
       }

        buttonFlag = 0;

        buttonLastState[0] = BUTTON0;
        buttonLastState[1] = BUTTON1;
        buttonLastState[2] = BUTTON2;
        buttonLastState[3] = BUTTON3;
        buttonLastState[4] = BUTTON4;

        if (buttonsPressed >= 5)
        {
            //0x00 00 00
            code = ((unsigned long)buttonCode[0] << 16) + ((unsigned long)buttonCode[1] << 12) +
                   ((unsigned long)buttonCode[2] << 8)  + ((unsigned long)buttonCode[3] << 4) + buttonCode[4];
            switch (code)
            {
                case 0x11111:           // test
                    setLED2(TEAL, GREEN);
                    break;

                case 0x13421:           // handshake
                    CYRF6936_wakeup();
                    makePacket(&tx_msg, ADDR_MASTER, FLAGS_NONE, CMD_SET_SEC_STATE, 0x10);
                    handshake_cmd[0] = CMD_SET_SEC_STATE;
                    returnAddr = sendNaligPacket(FREQ_A, &tx_msg, PAYLOAD_LEN_SM);  // send security toggle request
                    if (returnAddr)
                    {
                        setLED(OFF);
                        if (receiveNaligPacket(FREQ_A, PROCESS_PACKETS)) // receive and process challenge
                        {
                            if (!receiveNaligPacket(FREQ_A, PROCESS_PACKETS))    // receive and process access denied/granted message
                            {
                                setLED2(YELLOW, TEAL);
                            }
                        }
                        else
                        {
                            setLED2(YELLOW, BLUE);
                        }
                    } else
                    {
                        setLED(YELLOW);
                    }
                    CYRF6936_sleep();
                    break;

                case 0x12345:   // verify master is there
                    CYRF6936_wakeup();
                    makePacket(&tx_msg, ADDR_MASTER, FLAGS_NONE, CMD_TYPE_REQ, 0x00);
                    if(sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD-1))
                    {
                        setLED(PURPLE);
                        receiveNaligPacket(FREQ_A, PROCESS_PACKETS);
                    }
                    else
                    {
                        setLED(YELLOW);
                    }
                    CYRF6936_sleep();
                    break;
                case 0x44444:   // left garage door
                    num = 1;
                    goto open_garage_door;
                case 0x22222:   // right garage door
                    num = 2;
                    open_garage_door:
                    CYRF6936_wakeup();
                    makePacket(&tx_msg, 0xC0, FLAGS_NONE, CMD_PULSE_GPIO, num);
                    if(sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD)) {
                        setLED(GREEN);
                    } else {
                        setLED(YELLOW);
                    }
                    CYRF6936_sleep();
                    break;

                case 0x33333:   // play a sound
                    num = 100;
                    goto play_sound;
                case 0x31111:
                    num = 63;
                    goto play_sound;
                case 0x34444:
                    num = 53;
                    goto play_sound;
                case 0x32222:
                    num = 54;
                    goto play_sound;
                    
                    play_sound:
                    CYRF6936_wakeup();
                    makePacket(&tx_msg, TYPE_AUDIO, FLAG_BROADCAST + FLAG_RELAY, CMD_PLAY_AUDIO, num);
                    sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD);
                    setLED(PURPLE);
                    CYRF6936_sleep();
                    break;
                case 0x31313:
                    CYRF6936_wakeup();
                    makePacket(&tx_msg, TYPE_AUDIO, FLAG_BROADCAST, CMD_LOOP_AUDIO, 131);
                    sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD);
                    setLED(PURPLE);
                    CYRF6936_sleep();
                    break;
                default:
                    setLED(RED);
            }
            buttonsPressed = 0;
            T3CONbits.TON =0;
        } else if (!buttonUp) {
            setLEDfast(BLUE);
            TMR3 = 0;
            T3CONbits.TON = 1;
        }

        Idle();
   }
   */

}

void keyfobMainLoop()
{
   //int ret;
   //unsigned char returnAddr;

   if (buttonFlag)
   {
       // use timer 2 to delay for button debouncing
        T1CONbits.TON = 0;  // disable timer 1 so we don't sleep forever during debounce
        IEC1bits.CNIE = 0;  // disable input change interrupt
        T2CONbits.TON = 1;          // enable timer 2
        TMR2 = 0;
        Idle(); Nop();
        T2CONbits.TON = 0;          // disable timer 2
        IEC1bits.CNIE = 1;  // enable input change interrupt
        T1CONbits.TON = 1;

        // check thich buttons are pressed down
       if (BUTTON0 == 0)                // down
       {
            //setLED2(YELLOW, PURPLE);
//           if (badAddrFlag)
//           {
                makePacket(&tx_msg, ADDR_BROADCAST, FLAGS_NONE, CMD_ADDR_REQ, 0x00);   // make address assignment packet to the Temp Address
                if(sendNaligPacket(FREQ_A, &tx_msg, PAYLOAD_LEN_SM))
                {
                    if (!receiveNaligPacket(FREQ_A, PROCESS_PACKETS))
                    {
                        blinkLED(RED,3);
                    }
                }
                else
                {
                    setLED(RED);
                }
//           } else
//           {
//               setLED2(YELLOW, PURPLE);
//           }
       } else if (BUTTON1 == 0) {       // up
//            CYRF6936_wakeup();
//            makePacket(&tx_msg, ADDR_MASTER, FLAGS_NONE, CMD_TYPE_REQ, 0x00);
//            if(sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD-1))
//            {
//                setLED(PURPLE);
//                receiveNaligPacket(FREQ_A);
//            }
//            else
//            {
//                setLED(RED);
//            }
//            CYRF6936_sleep();
       } else if (BUTTON2 == 0) {       // right
            setLED(BLUE);
       } else if (BUTTON3 == 0) {       // left
            setLED2(TEAL, GREEN);
       } else if (BUTTON4 == 0 || tempFlag) {       // center
           tempFlag = 0;
//           CYRF6936_wakeup();
//            makePacket(&tx_msg, ADDR_MASTER, FLAGS_NONE, CMD_SET_SEC_STATE, 0x10);
//
//            T3CONbits.TON = 1;          // enable timer 3
//            TMR3 = 0;
//            handshake_cmd[0] = CMD_SET_SEC_STATE;
//            returnAddr = sendNaligPacket(FREQ_A, &tx_msg, PAYLOAD_LEN_SM);  // send security toggle request
//            if (returnAddr)
//            {
//                setLED(OFF);
//                if (receiveNaligPacket(FREQ_A)) // receive and process challenge
//                {
//                    if (!receiveNaligPacket(FREQ_A))    // receive and process access denied/granted message
//                    {
//                        setLED2(YELLOW, TEAL);
//                    }
//                }
//                else
//                {
//                    setLED2(YELLOW, BLUE);
//                    //blinkLED(BLUE, receivedPackets);
//
//                }
//                tempcount = TMR3;
////                ret = txHandshake();
////
////                switch (ret)
////                {
////                    case HANDSHAKE_CODE_GRANTED:    // code accepted
////                        //setLED(GREEN);  break;
////                        break;
////                    case HANDSHAKE_CODE_DENIED:
////                        //setLED(RED);  break;
////                        break;
////                    case HANDSHAKE_NO_ACK:
////                        setLED2(YELLOW, RED); break;
////                    case HANDSHAKE_NO_CHALLENEGE:
////                        setLED2(YELLOW, BLUE); break;
////                    case HANDSHAKE_NO_ACK_FROM_RESPONSE:
////                        setLED2(YELLOW, PURPLE); break;
////                    case HANDSHAKE_NO_STATUS_RETURNED:
////                        setLED2(YELLOW, TEAL); break;
////                    default:    // communication error during handshake
////                        setLED(YELLOW);
////                }
//            } else
//            {
//                setLED(RED);
//            }
//            T3CONbits.TON = 0;          // enable timer 3
//            CYRF6936_sleep();

       } else if (BUTTON0 && buttonLastState[0] == 0) { // down

       } else if (BUTTON1 && buttonLastState[1] == 0) { // up

       } else if (BUTTON2 && buttonLastState[2] == 0) { // right
           setLED(RED);
       } else if (BUTTON3 && buttonLastState[3] == 0) { // left

       } else if (BUTTON4 && buttonLastState[4] == 0) { // center

       }

        buttonFlag = 0;
        
        buttonLastState[0] = BUTTON0;
        buttonLastState[1] = BUTTON1;
        buttonLastState[2] = BUTTON2;
        buttonLastState[3] = BUTTON3;
        buttonLastState[4] = BUTTON4;

        
        Idle();
   }

   
}

void nodeMainLoop()
{
    int ret;
    unsigned char origBuffSize;
//    if (!badAddrFlag)
//    {
//        while (receiveNaligPacket(FREQ_A) );
//    } else {
//        ClrWdt();   // receiving a packet will involve idling, which resets the WDT.  if we aren't doing that, we need to clear it
//    }
    while (receiveNaligPacket(FREQ_A, PROCESS_PACKETS) );

    unsigned char bytesToSend,uartStartIndex=0;
    
    if (UART_RX_buffer_size > 0)    // do we have serial data?
    {
        //txChar(UART_RX_buffer[UART_RX_buffer_size - 1]);    // echo
        if (myConfig.myType == TYPE_SERIAL)  // if this is a serial node, we are just supposed to send the data through the UART
        {
            if (UART_RX_buffer_size == 3 && UART_RX_buffer[0] == '*')   // escape sequence out of serial mode: **^
            {
                if (UART_RX_buffer[1] == '*' && UART_RX_buffer[2] == '^') {
                    myConfig.myType = 0xFE;
                    return;
                }
            }
            origBuffSize = UART_RX_buffer_size;

            bytesToSend = UART_RX_buffer_size - uartStartIndex; // we can only send 9 bytes at a time, so we have to loop
            while (bytesToSend)
            {
                if (bytesToSend > 64) bytesToSend = 64;

                makePacket(&tx_msg, TYPE_SERIAL, FLAG_BROADCAST, CMD_SERIAL_DATA, 0x00);
                copyArray(&tx_msg.data[1], (unsigned char *)&UART_RX_buffer[uartStartIndex], bytesToSend);
                sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_MIN + 1 + bytesToSend);

                uartStartIndex += bytesToSend;
                rest(TIMEOUT_20MS);     // this delay helps avoid skipping characters
                bytesToSend = UART_RX_buffer_size - uartStartIndex;
            }
            setLED(TEAL);
            UART_RX_buffer_size = 0;
        }
        else if (UART_RX_buffer[UART_RX_buffer_size-1] == 13 || UART_RX_buffer_size >= UART_BUFFER_SIZE)
        {
            processSerialData();
            addRndBase( (rnd_adder & 0xFF));
            addRndBase( (rnd_adder & 0xFF00) >> 8);
        }
    }

    if (myConfig.myType == TYPE_MOTION && bSensorsEnabled)
    {
        if (GPIO1 == 0 && sensorDelay == 0)
        {
            rest(TIMEOUT_32MS);
            if (GPIO1 == 0)
            {
                makePacket(&tx_msg, ADDR_MASTER, FLAGS_NONE, CMD_REPORT_MOTION, 1);
                sendNaligPacket(FREQ_A, &tx_msg, PAYLOAD_LEN_SM);  
                TMR3 = 0;
                sensorDelay = 1;
                setLED(BLUE);
            }
        }
    } /* else if (myConfig.myType == TYPE_MOTIONX)
    {
        if (motionX_flag)
        {
            if (!bSensorsEnabled) return;
            makePacket(&tx_msg, ADDR_MASTER, FLAGS_NONE, CMD_REPORT_MOTION, 1);
            sendNaligPacket(FREQ_A, &tx_msg, PAYLOAD_LEN_SM+1);
            tx_msg.data[2] = last_motionX_counts;
            TMR3 = 0;
            setLED(BLUE);
            motionX_flag = 0;
        } 
    }*/ else if (myConfig.myType == TYPE_DOOR)
    {
        if (GPIO1 != doorSensorLast[0])
        {
            ret = GPIO1 ? CMD_DOOR_OPEN : CMD_DOOR_CLOSED;
            doorSensorLast[0] = GPIO1;
            if (!bSensorsEnabled) return;
            makePacket(&tx_msg, ADDR_MASTER, FLAGS_NONE, ret, 1);
            sendNaligPacket(FREQ_A, &tx_msg, PAYLOAD_LEN_SM);
            setLED(BLUE);
        }
        if (GPIO3 != doorSensorLast[1])
        {
            ret = GPIO3 ? CMD_DOOR_OPEN : CMD_DOOR_CLOSED;
            doorSensorLast[1] = GPIO3;
            makePacket(&tx_msg, ADDR_MASTER, FLAGS_NONE, ret, 3);
            sendNaligPacket(FREQ_A, &tx_msg, PAYLOAD_LEN_SM);
            setLED(BLUE);
        }
    }

    //if (tx_buffer_len) _U1TXInterrupt();
}

void processSerialData()
{
    //UART_RX_buffer[] contains serial data
    //UART_RX_buffer_size contains the number of bytes receieved
    //      There is an extra byte for the \r character used to delimit commands
    unsigned char type;
    unsigned char num, num2, num3, i, j=0;
    unsigned char dest_nodes[10], num_dest=0;
    //nalig_msg tx_msg;
    //int ret;
    
    switch (UART_RX_buffer[0])
    {
        case 'A':   // address assignment: A112233
            if (UART_RX_buffer_size != 8) goto bad_command; // must have 8 characters [1 char command + 2 char addy + 2 char hex address + 2 char hex type + \r]
            // convert the 2 ascii chars to a byte
            num = ASCIIhex(UART_RX_buffer[1], UART_RX_buffer[2]);   // current address of node
            num2 = ASCIIhex(UART_RX_buffer[3], UART_RX_buffer[4]);   // address to be assigned
            num3 = ASCIIhex(UART_RX_buffer[5], UART_RX_buffer[6]);   // device type to be assigned
            if (num2 == 0xFF || num3 == 0xFF || num2 == 0x00 || num3 == 0x00 ||num2 == 0xFE )  goto bad_command; // cannot make these assignments

            //assignAddr = num;
            //assignType = num2;

            makePacket(&tx_msg, num, FLAGS_NONE, CMD_ADDR_ASSIGN, num2);   // make address assignment packet
            tx_msg.data[2] = num3; // need another byte for the type
            if(sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD+1))
            {
                if (!receiveNaligPacket(FREQ_A, PROCESS_PACKETS))
                {
                    txStr("!\r");
                }
            }
            else
            {
                txStr("!\r");

            }

            //TMR3 = 0;           // they get 2 seconds to respond!
            // see CMD_ADDR_REQ in processPacket() for continuation of the address assignment
            break;

        case 'M':   // set my address M1122
            if (UART_RX_buffer_size != 6) goto bad_command; // must have 6 characters [1 char command + 2 char hex address + 2 char hex type + \r]
            // convert the 2 ascii chars to a byte
            num = ASCIIhex(UART_RX_buffer[1], UART_RX_buffer[2]);   // address to be assigned
            num2 = ASCIIhex(UART_RX_buffer[3], UART_RX_buffer[4]);   // device type to be assigned
            if (num == 0xFF || num2 == 0xFF || num == 0x00 || num2 == 0x00 ||num == 0xFE )  goto bad_command; // cannot make these assignments

            myConfig.myAddr = num;
            myConfig.myType = num2;
            saveDeviceConfig();
            
            _blinkLED(WHITE, 2);
            badAddrFlag = 0;
       case 'm':   // query address
            txReport(myConfig.myAddr, ":a", myConfig.myType);
            break;

        case 'T':   // query a node over RF
            if (UART_RX_buffer_size != 4) goto bad_command; // must have 4 characters [1 char command + 2 char hex address + \r]
            num = ASCIIhex(UART_RX_buffer[1], UART_RX_buffer[2]);   // address of node to talk to

            makePacket(&tx_msg, num, FLAGS_NONE, CMD_TYPE_REQ, 0x00);   // make type request packet
            if (sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD-1))
            {
                //txStr("*\r");
                if (receiveNaligPacket(FREQ_A, PROCESS_PACKETS))
                {
                    //txStr("*\r");
                } else {
                    txStr("!\r");
                }
            } else {
                txStr("!\r");
            }
            break;

        case 'R':    // send custom RF command: RXX1122334455...\r   (where XX is the destination, and 11, 22, etc are data bytes)
        case 'B':   // send custom RF command (broadcast flag enabled)
            if (UART_RX_buffer_size < 6) goto bad_command; // must have 6 characters [1 char command + 2 char hex address + 2 data chars + \r]
            if (UART_RX_buffer_size & 1) goto bad_command;    // we must have an even number of data bytes
            
            num = ASCIIhex(UART_RX_buffer[1], UART_RX_buffer[2]);   // address of node to talk to
            num2 = (UART_RX_buffer_size - 4) >> 1;         // number of data bytes
            num3 = 0;   // data byte counter
            
            makePacket(&tx_msg, num, FLAGS_NONE, 0, 0);   // make the packet
            for (i = 3; i < UART_RX_buffer_size-2; i = i + 2)
            {
                tx_msg.data[num3++] = ASCIIhex(UART_RX_buffer[i], UART_RX_buffer[i+1]);
            }
            
            if (tx_msg.data[0] == CMD_SET_SEC_STATE) handshake_cmd[0] = CMD_SET_SEC_STATE;
            if (UART_RX_buffer[0] == 'B') tx_msg.broadcast = 1;
            num = sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_MIN + num2);

            if (num)
            {
                txVal("* ", num);
            } else {
                txStr("!\r");
            }
            break;

        case 'r':   // send custom RF command to multiple recipients: RXXYY...,112233...\r (where XX, YY are destination nodes (max 10), and 11, 22 are data bytes)
            if (UART_RX_buffer_size < 7) goto bad_command; // must have 7 characters min [1 char command + 2 char hex address + comma + 2 data chars + \r]
            if ((UART_RX_buffer_size & 1) == 0) goto bad_command;    // we must have an odd number of data bytes

            for (i = 1; i < UART_RX_buffer_size-2; i++)
            {
                if (UART_RX_buffer[i] == ',') { 
                    num = i;
                    goto found_comma;
                }
            }
            goto bad_command;   // couldn't find a comma
            found_comma:
            if ((num & 1) == 0) goto bad_command;    // comma position should be odd
            num_dest = (num - 1) >> 1;              // number of destination nodes = comma position - 1 / 2 (each destination node is 2 bytes)
            if (num_dest > 10) goto bad_command;   // we can't send to more than 10 nodes (this is arbitrary)

            for (i = 1; i < num; i = i + 2) // fill our destination node array
            {
                dest_nodes[j++] = ASCIIhex(UART_RX_buffer[i], UART_RX_buffer[i+1]);
            }

            j = 0;
            makePacket(&tx_msg, dest_nodes[0], FLAGS_NONE, 0, 0);   // fill our data array
            for (i = num+1; i < UART_RX_buffer_size-2; i = i + 2)
            {
                tx_msg.data[j++] = ASCIIhex(UART_RX_buffer[i], UART_RX_buffer[i+1]);
            }

            if (tx_msg.data[0] == CMD_SET_SEC_STATE) handshake_cmd[0] = CMD_SET_SEC_STATE;
            //if (UART_RX_buffer[0] == 'B') tx_msg.broadcast = 1;

            for (i=0; i<num_dest; i++)
            {
                tx_msg.dest_addr = dest_nodes[i];
                //txArray((unsigned char *)&tx_msg,  PACKET_SIZE_MIN + j);
                num = sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_MIN + j);
                if (num)
                {
                    txVal("* ", tx_msg.dest_addr);
                } else {
                    txVal("! ", tx_msg.dest_addr);
                }
                increasePacket(&tx_msg);
                increaseSequence(&tx_msg);
            }
            i=i+1;
            break;

        case 'S':   // enable/disable sensors
            if (UART_RX_buffer_size != 3) goto bad_command; // must have 3 characters [1 char command + 1 char en/dis + \r]
            bSensorsEnabled = ASCIIdigit(UART_RX_buffer[1]) ? 1 : 0;
        case 's':
            txVal("Sensors: ", bSensorsEnabled);
            break;
        case 'i':
            if (UART_RX_buffer_size != 3) goto bad_command; // must have 3 characters [1 char command + 1 char en/dis + \r]
            ignore_non_relayed_packets = ASCIIdigit(UART_RX_buffer[1]) ? 1 : 0;
            txVal("Non-relay packet ignore: ", ignore_non_relayed_packets);
            break;
        case '$':   // get/set security state
             if (UART_RX_buffer_size == 3)  // set security state
             {
                  num = ASCIIdigit(UART_RX_buffer[1]);
                  if (num > 3) num = 3;
                  securityState = num;
                  txReport(0, ":S", securityState);  // report security state
             } else {
                if (myConfig.myType == TYPE_MASTER)  // get security state
                {
                    txVal(":$", securityState);
                } else {
                    txStr(":$!!\r");
                }
             }
            break;
        case 'H':
            makePacket(&tx_msg, ADDR_MASTER, FLAGS_NONE, CMD_SET_SEC_STATE, 0x10);
            handshake_cmd[0] = CMD_SET_SEC_STATE;
            num = sendNaligPacket(FREQ_A, &tx_msg, PAYLOAD_LEN_SM);  // send security toggle request
            if (num)
            {
                setLED(OFF);
                if (receiveNaligPacket(FREQ_A, PROCESS_PACKETS)) // receive and process challenge
                {
                    if (!receiveNaligPacket(FREQ_A, PROCESS_PACKETS))    // receive and process access denied/granted message
                    {
                        setLED2(YELLOW, TEAL);
                    }
                }
                else
                {
                    setLED2(YELLOW, BLUE);
                }
            } else
            {
                setLED(RED);
            }
            break;
        case 'Q':
            writeReg(XACT_CFG_ADR, 0x30);   // set mode to RX
            waitForSynth();
            readReg(RSSI_ADR);  // dummy register read
            txVal("RSSI: ", readReg(RSSI_ADR));
            break;
        case 'q':

            GPIO0 = 1;
            rest(TIMEOUT_10MS);
            rest(TIMEOUT_10MS);
            GPIO0 = 0;
            rest(TIMEOUT_20MS);
            GPIO1 = 1;
            rest(TIMEOUT_32MS);
            GPIO0 = 0;

            /*
            generateNonce(challenge);
            for (num=0; num<16; num++)
                txHex(challenge[num]);
            txChar('\r');*/
            break;
        case 'x':

            makePacket(&tx_msg, 0xFF, FLAGS_NONE, CMD_SPECIAL, 0x0A);   // make the packet
            sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD);
            //tx_msg.dest_addr = 0x31;
            sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD);
            
            txStr("ok!\r");
            break;

        case 'X':
            if (UART_RX_buffer_size != 4) goto bad_command; // must have 4 characters [1 char command + 2 char hex address + \r]
            num = ASCIIhex(UART_RX_buffer[1], UART_RX_buffer[2]);   // address of node to talk to

            makePacket(&tx_msg, num, FLAGS_NONE, 0x41, 0x00);   // make type experimental large packet
            for (num2 = 0; num2 <100; num2++)
                tx_msg.data[num2+1] = num2;
            if (sendNaligPacket(FREQ_A, &tx_msg, 123))
            {
                txStr("ok!\r");
            } else {
                txStr("fail!\r");
            }
            break;
        case 'I':   // change indicator LED
            if (UART_RX_buffer_size != 4) goto bad_command; // must have 4 characters [1 char command + 2 char hex error + \r]
            num = ASCIIhex(UART_RX_buffer[1], UART_RX_buffer[2]);   // address of node to talk to

            myConfig.LED_blink = num;
            saveDeviceConfig();
            break;

        case 'E':   // sensor reporting an error
            if (UART_RX_buffer_size != 4) goto bad_command; // must have 4 characters [1 char command + 2 char hex error + \r]
            num = ASCIIhex(UART_RX_buffer[1], UART_RX_buffer[2]);   // address of node to talk to

            makePacket(&tx_msg, lastDataRequestor, FLAGS_NONE, CMD_ERROR, num);   // make the packet
            sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD);
            txStr("*\r");
            break;

        case 'e':   // play audio track (broadcast to all audio nodes)
            if (UART_RX_buffer_size != 4) goto bad_command; // must have 4 characters [1 char command + 2 char hex + \r]
            num = ASCIIhex(UART_RX_buffer[1], UART_RX_buffer[2]);   // address of node to talk to

            makePacket(&tx_msg, TYPE_AUDIO, FLAG_BROADCAST, CMD_ENQ_AUDIO, num);   // make the packet
            sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD);
            break;
            
        case 'P':   // play audio track (broadcast to all audio nodes)
            if (UART_RX_buffer_size == 5) {             //  5 characters [1 char command + 3 digit track + \r]
                num = ASCIIdigit(UART_RX_buffer[1]);    //hundreds
                num2 = ASCIIdigit(UART_RX_buffer[2]);  // tens
                num3 = ASCIIdigit(UART_RX_buffer[3]);  //ones
            } else if (UART_RX_buffer_size == 4) {   //  4 characters [1 char command + 2 digit track + \r]
                num2 = ASCIIdigit(UART_RX_buffer[1]);    //hundreds
                num3 = ASCIIdigit(UART_RX_buffer[2]);  // tens
                num = 0;
            } else {
                goto bad_command;
            }
            i = (num * 100) + (num2 * 10) + num3;

            makePacket(&tx_msg, TYPE_AUDIO, FLAG_BROADCAST, CMD_PLAY_AUDIO, i);   // make the packet
            sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD);
            break;

        case 'p':   // stop audio track (broadcast to all audio nodes)
            makePacket(&tx_msg, TYPE_AUDIO, FLAG_BROADCAST, CMD_STOP_AUDIO, 0x00);   // make the packet
            sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD);
            break;

        case 'L':   // loop audio track (broadcast to all audio nodes)
            if (UART_RX_buffer_size != 4) goto bad_command; // must have 4 characters [1 char command + 2 char hex + \r]
            num = ASCIIhex(UART_RX_buffer[1], UART_RX_buffer[2]);   // address of node to talk to

            makePacket(&tx_msg, TYPE_AUDIO, FLAG_BROADCAST, CMD_LOOP_AUDIO, num);   // make the packet
            sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_CMD);
            break;

        case 'C':   // print packet cache
            printPacketCache();
            break;

        case 'k':   // acknowledgement from sensor
            //sensorACK=1;
            //GPIO0 = 0;
            break;

        case '>':   // report sensor data
            type = UART_RX_buffer[1];

            if (type == 'E') {
                if (UART_RX_buffer_size != 16) goto bad_command; // >E TT.T HH.H LL\r
            //} else {
            //    goto bad_command;
            }

            makePacket(&tx_msg, lastDataRequestor, FLAGS_NONE, CMD_SENSOR_REPORT, 0);   // make the packet
            copyArray(tx_msg.data+1, (unsigned char *)UART_RX_buffer, UART_RX_buffer_size-1) ;       // send the data
            sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_MIN+ UART_RX_buffer_size);

            break;
        case '~':   // send text
            if (UART_RX_buffer_size < 5) goto bad_command; // must have at least 5 chars
            num = ASCIIhex(UART_RX_buffer[1], UART_RX_buffer[2]);   // address of node to talk to

            makePacket(&tx_msg, num, FLAGS_NONE, CMD_SERIAL_DATA, 0);   // fill our data array
            j=1;
            for (i = 3; i < UART_RX_buffer_size-1; i++)
            {
                tx_msg.data[j++] = UART_RX_buffer[i];
            }
            tx_msg.data[j++] = '\r';
            num2 = sendNaligPacket(FREQ_A, &tx_msg, PACKET_SIZE_MIN+UART_RX_buffer_size-2);
            if (num2)
            {
                txVal("* ", num);
            } else {
                txVal("! ", num);
            }
        case '-':   // ignore any text starting with -
            UART_RX_buffer_size = 0;
            break;
        case '@':
            CYRF6936_init();
            txStr("-RF Reset\r");
            break;
        default:
            bad_command:
            //setLED(WHITE);
            //00txChar( '\r');
            txStr("!!!\r");
    }
    UART_RX_buffer_size = 0;
}

// ADC interrupt
/*
void __attribute__((interrupt, no_auto_psv)) _ADC1Interrupt(void)
{
    ADCresult = ADC1BUF0;
    IFS0bits.AD1IF = 0;     // clear ADC1 interrupt flag
}*/


// timer 1 interrupt: used for LED control (setLED, setLED2, blinkLED)
void  __attribute__((interrupt, no_auto_psv)) _T1Interrupt(void)
{
    
    //TMR1 = 0; //0xC000;v
    IFS0bits.T1IF = 0;
    timer1_flag = 1;
}

// timer 2 interrupt: used for rest() and sleepUntilIRQ() timing
void  __attribute__((interrupt, no_auto_psv)) _T2Interrupt(void)
{
    timer2_flag = 1;
    IFS0bits.T2IF = 0;
}

// timer 3 interrupt: used for long stuff like blink-alive LED
void  __attribute__((interrupt, no_auto_psv)) _T3Interrupt(void)
{
    IFS0bits.T3IF = 0;
    timer3_flag = 1;
    
    if (deviceIsKeyfob)     // timeout for a button sequence
    {
        T3CONbits.TON = 0;  // disable timer
        buttonsPressed = 0;
        menuLevel=0;
        setLED(RED);
    } else {    // standard node
        if (currentLED == OFF && myConfig.LED_blink)
        {
            setLED(PURPLE); // blink alive
        }

        sensorDelay = 0;
    }

}

// timer 4 interrupt: clear the cache after some time
void  __attribute__((interrupt, no_auto_psv)) _T4Interrupt(void)
{
    IFS1bits.T4IF = 0;
    timer4_flag = 1;
    /*
    if (motionX_counts > 1) {
        motionX_flag = 1;
    }
    last_motionX_counts = motionX_counts;
    motionX_counts = 0;
     * */
}

// timer 5 interrupt: used to make sure we don't transmit too fast
void  __attribute__((interrupt, no_auto_psv)) _T5Interrupt(void)
{
    IFS1bits.T5IF = 0;
    timer5_flag = 1;    // set flag
    T5CONbits.TON = 0;  // turn off timer
}

// input change notification interrupt
void __attribute__((interrupt, no_auto_psv)) _CNInterrupt (void)
{
    IFS1bits.CNIF = 0;
    PORTA;
    PORTB;
    
    if (IRQ) {
        IRQ_flag = 1;
    } else  {
        buttonFlag = 1;
    }

    /*if (myConfig.myType == TYPE_MOTIONX)
    {
        if (GPIO3 == 1)  motionX_counts++;
    }*/

    //TMR1 = 0;
}

// UART TX interrupt
void __attribute__ ((interrupt,no_auto_psv)) _U1TXInterrupt(void)
{
    IFS0bits.U1TXIF = 0;

    if (tx_buffer_len)
    {
        U1TXREG = tx_buffer[tx_buffer_pos++];
        if (tx_buffer_pos >= tx_buffer_len)
        {
            tx_buffer_pos = tx_buffer_len = tx_fifo = 0;
        }
    }/* else if (tx_fifo) {
        tx_fifo--;
    }*/
}

// UART RX interrupt - Occurs when a byte is receieved over serial
void __attribute__ ((interrupt,no_auto_psv)) _U1RXInterrupt(void)
{
    IFS0bits.U1RXIF = 0;    // clear interrupt flag
    while (U1STAbits.URXDA) // while the RX buffer has stuff in it
    {
       if (UART_RX_buffer_size >= UART_BUFFER_SIZE) {
           U1RXREG; // buffer is full
       } else {
        UART_RX_buffer[UART_RX_buffer_size++] = U1RXREG; // put it in our own buffer

        /*
        if ((UART_RX_buffer[UART_RX_buffer_size-1] == '\r') && (UART_RX_buffer[UART_RX_buffer_size-2] == 'k')) {
            sensorACK = 1;
            GPIO0 = 0;
        }*/

       }
    }
    rnd_adder = TMR2;
    if (U1STAbits.OERR) U1STAbits.OERR = 0; // clear buffer errors
}

void _blinkLED(unsigned char color, unsigned char num)
{
    blinkLED(color, num);
}

void _setLED(unsigned char color)
{
    setLED(color);
}