/* 
 * File:   main.h
 * Author: Daniel
 *
 * Created on January 10, 2013, 9:10 PM
 */

#ifndef MAIN_H
#define	MAIN_H

#ifdef	__cplusplus
extern "C" {
#endif

#define FIRMWARE_VER_MIN 0x0A
#define FIRMWARE_VER_MAJ 0x02
//#define DEVICE_IS_KEYFOB      // deprecated: this info is now stored in flash

#define LED_R   LATAbits.LATA4
#define LED_G   LATAbits.LATA0
#define LED_B   LATAbits.LATA1

#define PWR_ON      LATBbits.LATB2
#define PWR_ON_TRIS TRISBbits.TRISB2

#define RED     0b011
#define GREEN   0b110
#define BLUE    0b101
#define TEAL    0b100
#define PURPLE  0b001
#define YELLOW  0b010
#define WHITE   0b000
#define OFF     0b111

// for the keyfob
#define BUTTON0 PORTAbits.RA2
#define BUTTON1 PORTBbits.RB8
#define BUTTON2 PORTBbits.RB4
#define BUTTON3 PORTAbits.RA3
#define BUTTON4 PORTBbits.RB7   // TX port for nodes

#define GPIO0 PORTAbits.RA2
#define GPIO1 PORTBbits.RB8
#define GPIO2 PORTBbits.RB4
#define GPIO3 PORTAbits.RA3

#define GPIO0_LAT LATAbits.LATA2
#define GPIO1_LAT LATBbits.LATB8
#define GPIO2_LAT LATBbits.LATB4
#define GPIO3_LAT LATAbits.LATA3

#define GPIO0_TRIS TRISAbits.TRISA2
#define GPIO1_TRIS TRISBbits.TRISB8
#define GPIO2_TRIS TRISBbits.TRISB4
#define GPIO3_TRIS TRISAbits.TRISA3

#define GPIO0_PU CNPU2bits.CN30PUE
#define GPIO1_PU CNPU2bits.CN22PUE
#define GPIO2_PU CNPU1bits.CN1PUE
#define GPIO3_PU CNPU2bits.CN29PUE

#define IRQ PORTBbits.RB9

#define BUTTON_UNPRESSED -1
//#define BUTTON_PRESSED  5
//#define BUTTON_WAIT     2

// handshake return codes
#define HANDSHAKE_CODE_GRANTED 1
#define HANDSHAKE_CODE_DENIED 0
#define HANDSHAKE_NO_ACK -1
#define HANDSHAKE_NO_CHALLENEGE -2
#define HANDSHAKE_NO_ACK_FROM_RESPONSE -3
#define HANDSHAKE_NO_STATUS_RETURNED -4

#define UART_BUFFER_SIZE 64
#define TX_BUFFER_SIZE  64

#define DelayUs(x) delayCount = x >> 2; while(--delayCount){ Nop(); } // delay for x microseconds
#define LED_VAL(COLOR) LED_R = (COLOR & 0b100)>>2; LED_G = COLOR & 0b001; LED_B = (COLOR & 0b010)>>1;
#define setLED(COLOR) LED_VAL(COLOR); currentLED = COLOR;  T1CONbits.TON = 1; TMR1 = 0;
#define setLEDfast(COLOR) LED_VAL(COLOR); currentLED = COLOR;  nextLED = OFF; T1CONbits.TON = 1; TMR1 = 0xA000;
#define setLED2(COLOR, COLOR2) LED_VAL(COLOR); nextLED = COLOR2;  currentLED = COLOR;  T1CONbits.TON = 1; TMR1 = 0x4000;
#define blinkLED(COLOR, NUM) LED_R = LED_VAL(COLOR); blinkLED = COLOR; numBlinks = NUM-1; currentLED = COLOR;  T1CONbits.TON = 1;  TMR1 = 0x4000;
//#define setLED(COLOR) PORTA = ((COLOR & 0b100) << 2) + (COLOR & 0b11)

// convert two ascii bytes (hex characters) into an unsigned char
#define ASCIIhex(byte1, byte2) (((byte1 < 'A') ? byte1 - '0' : byte1 - 'A' + 10) << 4) +    \
                                ((byte2 < 'A') ? byte2 - '0' : byte2 - 'A' + 10)

#define ASCIIdigit(byte)    (byte - '0')

void keyfobMainLoopAdv();
void keyfobMainLoop();
void keyfobMainColors();
void nodeMainLoop();
int txHandshake();
int host_Handshake();
void processSerialData();
void _blinkLED(unsigned char color, unsigned char num);
void _setLED(unsigned char color);

void __attribute__ ((interrupt,no_auto_psv)) _U1TXInterrupt(void);
void __attribute__((interrupt, no_auto_psv)) _T1Interrupt(void);
void __attribute__((interrupt, no_auto_psv)) _CNInterrupt(void);

#ifdef	__cplusplus
}
#endif

#endif	/* MAIN_H */

