#include <xc.h>
#include "rf.h"
#include "main.h"
#include "AES.h"
#include "nalig_protocol.h"
#include "sys_functions.h"

extern sys_config myConfig;
extern unsigned char unsavedConfigChanges;
extern unsigned char deviceIsKeyfob;

unsigned char packetCounter=0;
unsigned int sequenceCounter=0;

struct cacheEntry packet_cache[PACKET_CACHE_SIZE];
struct cacheEntry *cacheFirst;   // the first entry in the linked list: most recent packet
struct cacheEntry *cacheNext;    // the next cache entry to place received packet data
struct cacheEntry *cacheOldest;  // the oldest valid cache entry
unsigned char cacheHasEntries = 0;

extern unsigned char assignType, assignAddr;
extern unsigned char currentPowerLevel;
extern unsigned char ignore_non_relayed_packets;

extern volatile unsigned char timer3_flag;

extern unsigned char rnd_base[16];
extern unsigned char rnd_counter;

nalig_msg* rx_nalig;

extern unsigned char badAddrFlag;
extern unsigned char lastDataRequestor;

unsigned char securityState=0;
unsigned char handshakeState;
extern unsigned char sensorACK;

// these arrays must be aligned to even addresses, or else the AES function will cause a device reset
//extern unsigned char my_key[16];
// **************************************
// CHANGE THIS KEY
// **************************************
unsigned char __attribute__((aligned(2))) my_key[16] = {0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,0x0F};
unsigned char __attribute__((aligned(2))) handshake_challenge[16];
unsigned char __attribute__((aligned(2))) handshake_valid_response[16];
unsigned char handshake_cmd[2];

unsigned char receivedPackets;
extern unsigned char bSensorsEnabled;


void makePacket(nalig_msg *packet, unsigned char dest_addr, unsigned char flags, unsigned char command, unsigned char data)
{
    packet->dest_addr = dest_addr;
    packet->source_addr = myConfig.myAddr;
    packet->broadcast = ((flags & FLAG_BROADCAST) != 0);
    packet->ack = ((flags & FLAGS_NONE) != 0);
    packet->relay = ((flags & FLAG_RELAY) != 0);
    packet->handshake = ((flags & FLAG_HANDSHAKE) != 0);
    packet->TTL = 0xE;
    packet->sequence_id = sequenceCounter++;
    packet->packet_id = packetCounter++;
    packet->data[0] = command;
    packet->data[1] = data;
}

void increaseSequence(nalig_msg *packet)
{
    packet->sequence_id = sequenceCounter++;
}

void increasePacket(nalig_msg *packet)
{
    packet->packet_id = sequenceCounter++;
}

void sendNaligAck(nalig_msg *rx, unsigned char freq)
{
    nalig_msg ackPacket;
    // sends a ack packet to the corresponding received packet
    // an ack packet has no command or data
    ackPacket.dest_addr = rx->source_addr;
    ackPacket.source_addr = myConfig.myAddr;
    ackPacket.broadcast = 0;
    ackPacket.ack = 1;
    ackPacket.relay = rx->relay;
    ackPacket.handshake = 0;
    ackPacket.TTL = 0xE;
    ackPacket.sequence_id = rx->sequence_id;
    ackPacket.packet_id = rx->packet_id;
    TX(freq, (unsigned char *)&ackPacket, PACKET_SIZE_ACK);
    //txStr("A\r");
}

unsigned char receiveNaligPacket(unsigned char freq, unsigned char doPacketProcess)
{
    // attempts to recevive a NaliG packet
    // returns 1 if a packet was recevied, 0 if not
    // when a packet is received and addressed to this node, it will send a Nalig Ack packet
    unsigned char packet_size;
    unsigned char cacheHitResult;
    //unsigned char rx_attempts=0;
    nalig_msg nalig_packet;

    
    try_again:
    packet_size = RX(freq, (unsigned char *)&nalig_packet, TIMEOUT_20MS);

    if (packet_size >= PACKET_SIZE_MIN)    // we will not process any packet that is not the minimum size
    {
        
        receivedPackets++;
        //rx_nalig = (nalig_msg*)rx_packet;
        //txStr("\rR ");
        rx_nalig = &nalig_packet;

        if (rx_nalig->source_addr == myConfig.myAddr)        // one of our own packets got back to us we don't need to do anything with it
        {
            goto try_again;
        }

        if (rx_nalig->TTL >= 0xE && ignore_non_relayed_packets) goto try_again; // used for debugging: will only "hear" packets that have been relayed (when enabled)

        TMR4=0;
        cacheHitResult = findPacketInCache(&nalig_packet);     // check if this packet is in the cache
        
        if (cacheHitResult == 0) {                               // we haven't seen this packet before
            addPacketToCache(&nalig_packet);
        } else if (cacheHitResult == 2) {                         // packet ID match, sequence ID match.
            goto try_again;                            // ignore, we have seen this same packet before (probably got it from a relay)
        }

        if ((nalig_packet.dest_addr == myConfig.myAddr && !nalig_packet.broadcast) )    // check if destination is my address (and not a broadcast type)
        {
            
analyze_packet:
            if (!nalig_packet.ack)                 // if this isn't an ACK packet, we should send one
                sendNaligAck(rx_nalig, freq);

            //if ((packet_size > PACKET_SIZE_MIN) && (!cacheHitResult) && doPacketProcess) // if it is big enough to be a command, and we haven't processed this packet ID already, process it
            if (packet_size > PACKET_SIZE_MIN) {
                if ((!cacheHitResult) && doPacketProcess)
                    processPacket(&nalig_packet, packet_size);
            }
            addRndBase(nalig_packet.packet_id);
            //clearArray(rx_packet,packet_size);

            return 1;
        }
        else if((nalig_packet.dest_addr == ADDR_BROADCAST) ||                       // check if it is a broadcast address
            (nalig_packet.dest_addr == myConfig.myType && nalig_packet.broadcast)  )     // check if it is a type broadcast to a matching type
        {
            SMALL_RND_DELAY;    // we aren't the only node that might respond, add a small delay
            goto analyze_packet;
        }
        else    // not my packet not my problem, that's what i say
        {
            if (nalig_packet.relay && nalig_packet.TTL && myConfig.relayEnabled) // is the relay flag set and is the TTL not expired?
            {
                nalig_packet.TTL--;
                SMALL_RND_DELAY;        // we aren't the only node that might respond, add a small delay
                TX(freq, (unsigned char *)&nalig_packet, packet_size);  // relay the packet
                //_setLED(BLUE);
            }
            //txStr("Ignoring Packet: ");
            //printPacket(rx_nalig);
            goto try_again;
        }
    } else if (packet_size > 0) {   // packet is too small, try again
        goto try_again;
    }

    //if (rx_attempts++ < 3) goto try_again;

    return 0;   // operation timed out, did not receieve any data
}

unsigned char sendNaligPacket(unsigned char freq, nalig_msg* packet, unsigned char length)
{
    // attempts to send a packet (without CYRF6936 auto-ack), but waits for a Nalig Ack packet
    // will not wait for an Ack if the packet is a broadcast
    // returns 0 if no ack received, returns the address of the acknowledger if success
    unsigned char cycles=0;

    if (deviceIsKeyfob) {
        if (currentPowerLevel > RF_MIN_PWR) currentPowerLevel--;    // try reducing power if possible!
    } else {
        currentPowerLevel = RF_MAX_PWR;
    }

    while (cycles < 4)
    {
        if (TX(freq, (unsigned char *)packet, length) != RF_TX_FAIL)    // transmit the packet
        {
            //if (packet->broadcast || packet->dest_addr == ADDR_BROADCAST) return 0xFF;      // don't wait for an ack on a broadcast
            if (receiveNaligPacket(freq, DO_NOT_PROCESS_PACKETS))  // try to receieve a packet (we are looking for an ack)
            {
                if (rx_nalig->ack && (rx_nalig->packet_id == packet->packet_id))    // did we get an appropriate ack?
                    return rx_nalig->source_addr;    // return the responder's address
            }
        } // got bad packet or incorrect ack
        if (currentPowerLevel < RF_MAX_PWR) currentPowerLevel++;    // increase the power if we can
        packet->relay = 1;          // allow the packet to be relayed
        increaseSequence(packet);   // increase the sequence number
        cycles++;
    }
    return 0;
}

unsigned char sendNaligPacketMulti(unsigned char freq, nalig_msg* packet, unsigned char length, unsigned char numACKs)
{
    // attempts to send a packet (without CYRF6936 auto-ack), but waits for MULTIPLE Nalig Ack packet
    // will rebroadcast the same packet until X unique acks are received, or it "times out"
    // returns the number of acks received (up to numACKs)
    // num ACKs must be < 10
    
    unsigned char cycles=0;
    unsigned char acks[10], ackCount=0;

    if (deviceIsKeyfob) {
        if (currentPowerLevel > RF_MIN_PWR) currentPowerLevel--;    // try reducing power if possible!
    } else {
        currentPowerLevel = RF_MAX_PWR;
    }

    while (cycles < 4)
    {
        if (TX(freq, (unsigned char *)packet, length) != RF_TX_FAIL)    // transmit the packet
        {
            listen_again:
            if (receiveNaligPacket(freq, DO_NOT_PROCESS_PACKETS))  // try to receieve a packet (we are looking for an ack)
            {
                if (rx_nalig->ack && (rx_nalig->packet_id == packet->packet_id)) {   // did we get an appropriate ack?
                    if (!inArray(rx_nalig->source_addr, acks, ackCount)) // have we seen this sender before?
                    {
                        acks[ackCount++] = rx_nalig->source_addr;   // it's new, add it to the list
                    }
                    if (ackCount >= numACKs) {
                        //txArray(acks, ackCount);
                        return ackCount;    // have we seen all the acks we need?
                    }
                    goto listen_again;
                }
            }
        } // got bad packet or incorrect ack
        if (currentPowerLevel < RF_MAX_PWR) currentPowerLevel++;    // increase the power if we can
        packet->relay = 1;          // allow the packet to be relayed
        increaseSequence(packet);   // increase the sequence number
        cycles++;
    }
    
    return ackCount;
}


unsigned char processPacket(nalig_msg *packet, unsigned char num_bytes)
{
    // num_bytes is the total size of the packet (including all header bytes), not just the data section
    unsigned char num, num2, cmd, time;
    unsigned char error_code;
    unsigned char ret;
    nalig_msg msg;

    switch (packet->data[0])
    {
        case CMD_ERROR: // error
            _blinkLED(YELLOW, 2);
            txChar('<'); txHex(packet->source_addr);
            txVal("!", packet->data[1]);
            break;
        case CMD_PING:      // Ping
            if (packet->data[1] == 0x00)
            {
                makePacket(&msg, packet->source_addr, FLAGS_NONE, CMD_PING, 0x01);
                msg.relay = packet->relay;  // if we got a relayed packet, our response will probably also need to be relayed
                TX(FREQ_A, (unsigned char *)&msg, PACKET_SIZE_CMD);
            }
            break;
        case CMD_ADDR_ASSIGN:  // address assignment
            if (GPIO0 || myConfig.myType == TYPE_MASTER || packet->broadcast) {
                _setLED(YELLOW);
                goto reject_new_addr;    // are we expecting a request?  no? then fuck off!
            }
            num = packet->data[1];
            num2 = packet->data[2];
            if (num == 0xFF || num2 == 0xFF || num == 0x00 || num2 == 0x00 ||num == 0xFE )
            {
                goto reject_new_addr;
            }

            myConfig.myAddr = num;
            myConfig.myType = num2;
            saveDeviceConfig();
            makePacket(&msg, packet->source_addr, FLAGS_NONE, CMD_ADDR_RESP, 0x01);
            sendNaligPacket(FREQ_A, &msg, PACKET_SIZE_CMD);
            _blinkLED(WHITE, 2);
            txReport(myConfig.myAddr, ";a", myConfig.myType);
            badAddrFlag = 0;
            break;

            reject_new_addr:
                makePacket(&msg, packet->source_addr, FLAGS_NONE, CMD_ADDR_RESP, 0x00);
                sendNaligPacket(FREQ_A, &msg, PACKET_SIZE_CMD);
            break;

        case CMD_ADDR_RESP: // address assignment response
            if(packet->data[1] == 0x01)
            {
                txStr(":AS\r");
            }
            else
            {
                txStr(":AF\r");
            }
            break;


        case CMD_STATUS:    // device reporting a node status
            txReport(packet->source_addr, ":N", packet->data[1]);
            break;

        case CMD_TYPE_REQ:
            makePacket(&msg, packet->source_addr, FLAGS_NONE, CMD_TYPE_RESP, myConfig.myType);   // make address assignment packet to the Temp Address
            msg.data[2] = FIRMWARE_VER_MAJ;
            msg.data[3] = FIRMWARE_VER_MIN;
            msg.relay = packet->relay;  // if we got a relayed packet, our response will probably also need to be relayed
            sendNaligPacket(FREQ_A, &msg, PACKET_SIZE_CMD+2);
            break;

        case CMD_TYPE_RESP:
            txReport2(packet->source_addr, ":T", packet->data[1], packet->data[2]);
            break;

        case CMD_DEVICE_CFG:
            if (packet->data[1] & 0x80) {   // this is a SET DEVICE CONFIG command
                if (num_bytes < PACKET_SIZE_CMD+1) {    // must have 3 data bytes
                    error_code = ERR_CMD_TOO_SMALL;
                    goto bad_rf_command;
                }
                switch (packet->data[1] & 0x7F)
                {
                    case 0x01:  myConfig.LED_blink = packet->data[2] & 0x01;
                                _setLED((packet->data[2] & 0x0E) >> 1);
                                saveDeviceConfig();  break;
                    case 0x02:  myConfig.relayEnabled = packet->data[2];
                                saveDeviceConfig();  break;
                    case 0x03:  bSensorsEnabled = packet->data[2];  break;
                    case 0x04:  myConfig.hasRFE = packet->data[2];
                                saveDeviceConfig();  break;
                    default:    error_code = ERR_INVALID_PARAM;
                                goto bad_rf_command;
                }
                txVal2("Config set: ", packet->data[1], packet->data[2]);
            } else {                        // this is a GET DEVICE CONFIG command
                switch (packet->data[1] & 0x7F)
                {
                    case 0x01:  num = myConfig.LED_blink;       break;
                    case 0x02:  num = myConfig.relayEnabled;     break;
                    case 0x03:  num = bSensorsEnabled;  break;
                    case 0x04:  num = myConfig.hasRFE;  break;
                    default:    error_code = ERR_INVALID_PARAM;
                                goto bad_rf_command;
                }
                makePacket(&msg, packet->source_addr, FLAGS_NONE, CMD_REPORT_CFG, packet->data[1]);
                msg.data[2] = num;
                msg.relay = packet->relay;  // if we got a relayed packet, our response will probably also need to be relayed
                sendNaligPacket(FREQ_A, &msg, PACKET_SIZE_CMD+1);
            }
            break;

        case CMD_REPORT_CFG:
            txVal2("Config get: ", packet->data[1], packet->data[2]);
            break;

        case CMD_SET_SEC_STATE: // set security state
            if (myConfig.myType != TYPE_MASTER) {
                error_code = ERR_INVALID_CMD_CONTEXT;
                goto bad_rf_command;
            }

            if (packet->data[1] == 0xF0)    // just requesting the security status
            {
                makePacket(&msg, packet->source_addr, FLAGS_NONE, CMD_SEC_STATE, securityState);
                msg.relay = packet->relay;  // if we got a relayed packet, our response will probably also need to be relayed
                ret = sendNaligPacket(FREQ_A, &msg, PACKET_SIZE_CMD);
                break;
            }

            generateNonce(handshake_challenge);   // generate our pseudorandom 16 byte challenge

            makePacket(&msg, packet->source_addr, FLAGS_NONE, CMD_HS_CHALLANGE, 0x00);  // prepare a HANDSHAKE CHALLENGE packet
            msg.relay = packet->relay;  // if we got a relayed packet, our response will probably also need to be relayed
            copyArray(msg.data+1, handshake_challenge, 16);                 // copy the challenge data into the data packet

            copyArray(handshake_valid_response, handshake_challenge, 16);   // copy the challenge data for encryption
            AESEncrypt((int*)handshake_valid_response, (int*)my_key);       // encrypt the challenge ourselves to determine the correct response

            if(sendNaligPacket(FREQ_A, &msg, PACKET_SIZE_HS))
            {
                handshake_cmd[0] = packet->data[0];
                handshake_cmd[1] = packet->data[1];
                return packet->data[0];
            }
            else
            {
                handshake_cmd[0] = 0;
                txStr("HS: No challenge ack.\r");
            }
            break;

        case CMD_HS_CHALLANGE:
            if (num_bytes < PACKET_SIZE_HS) {
                error_code = ERR_CMD_TOO_SMALL;
                goto bad_rf_command;
            }
            copyArray(handshake_challenge, packet->data+1, 16);             // copy the challenge
            AESEncrypt((int*)handshake_challenge, (int*)my_key);            // compute the response

            makePacket(&msg, packet->source_addr, FLAGS_NONE, CMD_HS_RESPONSE, 0x00);   // prepare a HANDSHAKE RESPONSE packet
            msg.relay = packet->relay;  // if we got a relayed packet, our response will probably also need to be relayed
            copyArray(msg.data+1, handshake_challenge, 16);                 // put response into data payload
            ret = sendNaligPacket(FREQ_A, &msg, PACKET_SIZE_MIN+17);        // send 23 bytes (6 byte header + 1 byte command + 16 byte response)
            break;

        case CMD_HS_RESPONSE:
            if (num_bytes < PACKET_SIZE_HS) {   // did we get all 16 bytes of the response?
                error_code = ERR_CMD_TOO_SMALL;
                goto bad_rf_command;
            }
            if (!handshake_cmd[0]) {    // we have to be expecting this
                error_code = ERR_INVALID_CMD_CONTEXT;
                goto bad_rf_command;
            }
            if (compareArrays(handshake_valid_response, packet->data+1, 16))  // compare received response to a valid one
            {
                if (handshake_cmd[1] < 4)
                {
                    securityState = handshake_cmd[1];
                } else if (handshake_cmd[1] == 0x10)
                {
                    securityState = (securityState > 0) ? 0 : 3;    // togle security state between 0 and 3
                }
                makePacket(&msg, packet->source_addr, FLAGS_NONE, CMD_SEC_STATE, securityState);
                msg.relay = packet->relay;  // if we got a relayed packet, our response will probably also need to be relayed
                ret = sendNaligPacket(FREQ_A, &msg, PACKET_SIZE_CMD);

                txReport(packet->source_addr, ":S", securityState);  // Security access Granted
                _setLED(GREEN);
            } else {
                makePacket(&msg, packet->source_addr, FLAGS_NONE, CMD_ACCESS_DENIED, 0x00);
                ret = sendNaligPacket(FREQ_A, &msg, PACKET_SIZE_CMD);

                txVal(":s", packet->source_addr);  // Security access Denied
            }
            handshake_cmd[0] = 0;
            handshake_cmd[1] = 0;

            break;

        case CMD_ACCESS_DENIED: //access denied message
            txVal("Access denied: ", packet->source_addr);
            _blinkLED(RED, 2);

            break;
        case CMD_SEC_STATE:     // current security state (access granted)
            if (packet->data[1] > 0)
            {
                _blinkLED(GREEN, 2);
            } else {
                _setLED(GREEN);
            }
            break;

        case CMD_REPORT_GPIO:
            txReport(packet->source_addr, ":G", packet->data[1]);

            break;

        case CMD_SET_GPIO:
            if (num_bytes < PACKET_SIZE_CMD+1) {    // must have 3 data bytes
                error_code = ERR_CMD_TOO_SMALL;
                goto bad_rf_command;
            } else if (deviceIsKeyfob) {    // you can't set the GPIO of a keyfob
                error_code = ERR_INVALID_CMD_CONTEXT;
                goto bad_rf_command;
            }
            setGPIO(packet->data[1],packet->data[2]);
            saveDeviceConfig();
            break;

        case CMD_GET_GPIO:
            // respond with the current state of the GPIO pins
            num = GPIO0 + (GPIO1 << 1) + (GPIO2 << 2) + (GPIO3 << 3);
            //num = GPIO0;
            makePacket(&msg, packet->source_addr, FLAGS_NONE, CMD_REPORT_GPIO, num);
            msg.relay = packet->relay;  // if we got a relayed packet, our response will probably also need to be relayed
            sendNaligPacket(FREQ_A, &msg, PACKET_SIZE_CMD);
            break;

        case CMD_PULSE_GPIO:    // toggle a GPIO pin (and force as an output)
            num = packet->data[1] > 3 ? 3 : packet->data[1];
            time =  packet->data[2] == 0 ? 5 : packet->data[2];
            if (num == 0) { GPIO0_TRIS = 0; GPIO0 = 1; }
            else if (num == 1) { GPIO1_TRIS = 0; GPIO1 = 1; }
            else if (num == 2) { GPIO2_TRIS = 0; GPIO2 = 1; }
            else if (num == 3) { GPIO3_TRIS = 0; GPIO3 = 1; }

            // being super lazy here
            // delay 100 ms, N times, where N is the value from data[2]
            for (num2 = 0; num2 < time; num2++ )
            {
                rest(TIMEOUT_20MS); rest(TIMEOUT_20MS);
                rest(TIMEOUT_20MS); rest(TIMEOUT_20MS);
                rest(TIMEOUT_20MS);
            }
            /*
            timer3_flag = 0;
            T3CONbits.TON = 1;
            TMR3 = 0x8000;
            while (!timer3_flag) { }
            */
            
            if (num == 0) {  GPIO0 = 0; }
            else if (num == 1) { GPIO1 = 0; }
            else if (num == 2) { GPIO2 = 0; }
            else if (num == 3) { GPIO3 = 0; }
            break;
            
        case CMD_REPORT_MOTION:
            if (myConfig.myType != TYPE_MASTER) goto bad_rf_command;
            txReport(packet->source_addr, ":M", packet->data[1]);
            break;

        case CMD_DOOR_OPEN:
            if (myConfig.myType != TYPE_MASTER) goto bad_rf_command;
            txReport(packet->source_addr, ":D", packet->data[1]);
            break;

        case CMD_DOOR_CLOSED:
            if (myConfig.myType != TYPE_MASTER) goto bad_rf_command;
            txReport(packet->source_addr, ":d", packet->data[1]);
            break;

        case CMD_READ_SENSOR:   // read data from a serial-based sensor
            if (myConfig.myType == TYPE_TEMP)   // environode
            {
                cmd = 'a';
            } else {    // we aren't a serial-based sensor!!
                error_code = ERR_INVALID_CMD_CONTEXT;
                goto bad_rf_command;
            }

            lastDataRequestor = packet->source_addr;    // so we know who to send the result back to
            num = sensorACK=0;  // reset flags
            while (num++ < MAX_SENSOR_CMD_ATTEMPTS && !sensorACK)   // try a few times to see if we get a response
            {
                txChar(cmd); txChar('\r');
                sensorACK = waitForSerialAck(TIMEOUT_32MS);
            }
            // if we got an acknowledgement, the sensor will send serial data back and other code will process it later
            goto checkSerialSensorAck;
            break;

        case CMD_SENSOR_REPORT:
            txChar('<'); txHex(packet->source_addr);
            num2 = num_bytes - 1 - PACKET_SIZE_MIN; // calculate the number of data bytes
            for (num=0; num<num2; num++)
            {
                txChar(packet->data[num+1]);  // transmit the data
            }
            txChar('\r');
            break;

        case CMD_STOP_AUDIO:
            if (myConfig.myType != TYPE_AUDIO) {
                error_code = ERR_INVALID_CMD_CONTEXT;
                goto bad_rf_command;
            }
            lastDataRequestor = packet->source_addr;    // so we know who to send the result back to
            num = sensorACK=0;  // reset flags
            while (num++ < MAX_SENSOR_CMD_ATTEMPTS && !sensorACK)   // try a few times to see if we get a response
            {
                txStr("S\r");
                sensorACK = waitForSerialAck(TIMEOUT_32MS);
            }
            // if we got an acknowledgement, the sensor will send serial data back and other code will process it later
            goto checkSerialSensorAck;
            break;

        case CMD_PLAY_AUDIO:
            if (myConfig.myType != TYPE_AUDIO) {
                error_code = ERR_INVALID_CMD_CONTEXT;
                goto bad_rf_command;
            }
            lastDataRequestor = packet->source_addr;    // so we know who to send the result back to
            num = sensorACK=0;  // reset flags
            while (num++ < MAX_SENSOR_CMD_ATTEMPTS && !sensorACK)   // try a few times to see if we get a response
            {
                txVal("P", packet->data[1]);
                sensorACK = waitForSerialAck(TIMEOUT_32MS);
            }
            // if we got an acknowledgement, the sensor will send serial data back and other code will process it later
            goto checkSerialSensorAck;
            break;

        case CMD_ENQ_AUDIO:
            if (myConfig.myType != TYPE_AUDIO) {
                error_code = ERR_INVALID_CMD_CONTEXT;
                goto bad_rf_command;
            }
            lastDataRequestor = packet->source_addr;    // so we know who to send the result back to
            num = sensorACK=0;  // reset flags
            while (num++ < MAX_SENSOR_CMD_ATTEMPTS && !sensorACK)   // try a few times to see if we get a response
            {
                txVal("E", packet->data[1]);
                sensorACK = waitForSerialAck(TIMEOUT_32MS);
            }
            // if we got an acknowledgement, the sensor will send serial data back and other code will process it later
            goto checkSerialSensorAck;
            break;

        case CMD_LOOP_AUDIO:
            if (myConfig.myType != TYPE_AUDIO) {
                error_code = ERR_INVALID_CMD_CONTEXT;
                goto bad_rf_command;
            }
            lastDataRequestor = packet->source_addr;    // so we know who to send the result back to
            num = sensorACK=0;  // reset flags
            while (num++ < MAX_SENSOR_CMD_ATTEMPTS && !sensorACK)   // try a few times to see if we get a response
            {
                txVal("L", packet->data[1]);
                sensorACK = waitForSerialAck(TIMEOUT_32MS);
            }
            // if we got an acknowledgement, the sensor will send serial data back and other code will process it later
            goto checkSerialSensorAck;
            break;

        case CMD_SET_AUDIO_VOL:
            if (myConfig.myType != TYPE_AUDIO) {
                error_code = ERR_INVALID_CMD_CONTEXT;
                goto bad_rf_command;
            }
            lastDataRequestor = packet->source_addr;    // so we know who to send the result back to
            num = sensorACK=0;  // reset flags
            while (num++ < MAX_SENSOR_CMD_ATTEMPTS && !sensorACK)   // try a few times to see if we get a response
            {
                txVal("V", packet->data[1]);
                sensorACK = waitForSerialAck(TIMEOUT_32MS);
            }
            // if we got an acknowledgement, the sensor will send serial data back and other code will process it later
            goto checkSerialSensorAck;
            break;

        // ---------------------------------------------------------------
            checkSerialSensorAck:
            if (!sensorACK) {
                error_code = ERR_NO_SERIAL_RESPONSE;
                goto bad_rf_command;
            }
            break;
        // ---------------------------------------------------------------

        case CMD_SERIAL_DATA:
            //if (myConfig.myType != TYPE_SERIAL) {
             //   error_code = ERR_INVALID_CMD_CONTEXT;
             //   goto bad_rf_command;
            //}
            if (myConfig.myType == TYPE_SERIAL) // only blink if it's a serial node
            {
                _setLED(BLUE);
            }
            num2 = num_bytes - 1 - PACKET_SIZE_MIN; // calculate the number of data bytes
            
            for (num=0; num<num2; num++)
            {
                while ( U1STAbits.UTXBF || U1STAbits.TRMT == 0);   // wait while Tx buffer full
                U1TXREG = packet->data[num+1];  // transmit the data
            }

            break;

        case 0x41:
            num2 = num_bytes - 1 - PACKET_SIZE_MIN; // calculate the number of data bytes
            //txArray(packet->data+1, num2);
            for (ret=1; ret<num2; ret++)
                txHex(packet->data[ret]);
            txChar('\r');
            //txVal("Got bytes: ", num_bytes);
            break;

        case CMD_SPECIAL:
            txVal2("%", packet->data[1], packet->source_addr);
            break;

        default:    // invalid / unknown command
            error_code = ERR_INVALID_CMD;
            bad_rf_command:
            if (packet->dest_addr == myConfig.myAddr)    // only respond if this was sent directly to us
            {
                makePacket(&msg, packet->source_addr, FLAGS_NONE, CMD_ERROR, error_code);
                TX(FREQ_A, (unsigned char *)&msg, PACKET_SIZE_CMD);

                txStr(":!"); txHex(packet->source_addr);
                txChar('!'); txHex(error_code);
                txVal(",", packet->data[0]);
                _blinkLED(RED, 2);
            }


    }
    //printPacket(packet);

    return packet->data[0];
}


void initCache()
{
    // initialize all cache entries as invalid, and create the linked list
    // the next free entry will be the first entry
    unsigned char i;
    packet_cache[0].prev = (0);
    packet_cache[0].next = &packet_cache[1];
    packet_cache[0].valid = 0;
    cacheHasEntries = 0;
    for (i=1; i<PACKET_CACHE_SIZE-1; i++)
    {
        packet_cache[i].next = &packet_cache[i+1];
        packet_cache[i].prev = &packet_cache[i-1];
        packet_cache[i].valid = 0;
    }
    packet_cache[PACKET_CACHE_SIZE-1].prev = &packet_cache[PACKET_CACHE_SIZE-2];
    packet_cache[PACKET_CACHE_SIZE-1].next = (0);
    packet_cache[PACKET_CACHE_SIZE-1].valid = 0;
    cacheOldest = cacheNext = cacheFirst = &packet_cache[0];

}

unsigned char findPacketInCache(nalig_msg *packet)
{
    // traverse the linked-list cache to find a matching packet
    // return 0 if no match was found
    // return 1 if matching source, dest and packet ID
    // return 2 if matching source, dest, packet ID and sequence ID
    // if any match is found, move it to the front of the list
    struct cacheEntry *entry;

    entry = cacheFirst;      // start at the first (most recent) cache entry
    while (entry->valid)
    {
        if (packet->source_addr == entry->source_addr &&   // do we have a match?
            packet->dest_addr == entry->dest_addr &&
            packet->packet_id == entry->packet_id )
        {    // we have a match!

            if (entry->prev)    // if it's not the first entry, it needs to be
            {
                if (entry == cacheOldest)            // is this the oldest cache entry?
                   cacheOldest = entry->prev;       // now the 2nd oldest is the oldest
                entry->prev->next = entry->next;    // remove entry from the linked list
                if (entry->next)                    // do we have entries after this?
                   entry->next->prev = entry->prev;  // link the previous's next to this one's next
                entry->next = cacheFirst;            // make the 1st entry now the 2nd
                entry->prev = (0);                   // first entry doesn't have a previous

                cacheFirst->prev = entry;           // make the 2nd entry's previous be the first
                cacheFirst = entry;                 // this entry is now the first
            }

            if (packet->sequence_id == entry->sequence_id)
                return 2;           // sequence ID match and packet ID match
            else
                return 1;           // just a packet ID match (no sequence match)
        }
        if (entry->next == (0)) return 0;
        if (entry->next == cacheFirst) return 0;     // do i need this line?
        entry = entry->next;           // go to the next entry in the cache
    }

    return 0;
}

void addPacketToCache(nalig_msg *packet)
{
    // adds a packet to the cache
    // if the cache isn't full, it will go in the next available entry
    // if the cache is full, it will replace the oldest cache entry
    struct cacheEntry *entry;

    cacheHasEntries = 1;
    entry = cacheNext;
    entry->source_addr = packet->source_addr;
    entry->dest_addr = packet->dest_addr;
    entry->packet_id = packet->packet_id;
    entry->sequence_id = packet->sequence_id;
    entry->valid = 1;

    // point next cacheNext to free cache entry or next oldest item
    if (cacheNext->next == (0)) {        // are we at the end of the list? no more free entries after this one
        if (cacheNext == cacheOldest) {    // have we completely filled the list?
           cacheOldest = cacheNext = cacheOldest->prev;   // the next entry will replace the 2nd oldest
        } else {
           cacheNext = cacheOldest;         // the next entry will replace the oldest (this basically only happens once, the first time the cache is filled)
        }
        cacheOldest->next = (0);           // no entries after the last one!
    } else {
        cacheNext = cacheNext->next;    // still have more unused cache entries
    }

    if (entry->prev)    // if it's not the first entry, it needs to be
    {
        cacheFirst->prev = entry;           // put entry at the front
        entry->prev = (0);                  // the first entry doesn't have a previous entry
        entry->next = cacheFirst;           // the previously 1st entry is now the 2nd entry
    } else {        // first time adding an entry to the cache
        entry->next = (0);   // none follow, this is the first and only entry
    }
    cacheFirst = entry;   // this entry is now the new first
}

void printPacketCache()
{
    // print the cache - for debugging purposes only
    struct cacheEntry *entry;
    unsigned char i=0;
    
    txStr("Cache:\r");
    entry = cacheFirst;
    while (entry->valid)
    {
        txHex(i); txStr(": ");
        txHex(entry->source_addr); txChar(' ');
        txHex(entry->dest_addr); txChar(' ');
        txHex(entry->packet_id); txChar(' ');
        txInt(entry->sequence_id); txChar('\r');

        if (entry->next == (0)) goto finish;
        entry = entry->next;
        i++;
    }

    finish:
    txStr("Done.\r");
}