/*
 * File:   nalig_protocol.h
 * Author: Daniel
 *
 * Created on March 10, 2013, 4:16 PM
 */

#ifndef NALIG_PROTOCOL_H
#define	NALIG_PROTOCOL_H

#ifdef	__cplusplus
extern "C" {
#endif


#define ADDR_BROADCAST  0xFF        // broadcast address (all nodes will process)
#define ADDR_TEMP       0xFE        // temp address used when we don't have a real address
#define ADDR_MASTER     0x01        // address of the master node

#define TYPE_MASTER     0x01
#define TYPE_KEYFOB     0x02
#define TYPE_AUDIO      0x03
#define TYPE_MOTION     0x04
#define TYPE_DOOR       0x05
#define TYPE_MOTIONX    0x06
#define TYPE_LIGHT      0x07
#define TYPE_TEMP       0x08
#define TYPE_SERIAL     0x10

#define FLAG_BROADCAST  8
#define FLAG_ACK        4
#define FLAG_RELAY      2
#define FLAG_HANDSHAKE  1
#define FLAGS_NONE      0

#define STATUS_ONLINE       0x01
#define STATUS_RADIO_TEST   0x02
#define STATUS_ALERT        0x03

#define CMD_ERROR           0x00        // last command received was an error
#define CMD_ADDR_ASSIGN     0x01        // assigning a new adress and type
#define CMD_ADDR_REQ        0x02        // requesting an address and type
#define CMD_ADDR_RESP       0x03
#define CMD_STATUS          0x04        // status notification
#define CMD_PING            0x05        // ping to see if a node is still there
#define CMD_TYPE_REQ        0x06
#define CMD_TYPE_RESP       0x07
#define CMD_DEVICE_CFG      0x08
#define CMD_REPORT_CFG      0x09
    
#define CMD_SPECIAL         0x0F

#define CMD_SET_SEC_STATE   0x10
#define CMD_SEC_STATE       0x11
#define CMD_ACCESS_DENIED   0x12
#define CMD_HS_CHALLANGE    0x13
#define CMD_HS_RESPONSE     0x14

#define CMD_REPORT_GPIO     0x20
#define CMD_SET_GPIO        0x21
#define CMD_GET_GPIO        0x22
#define CMD_PULSE_GPIO      0x23
    
#define CMD_REPORT_MOTION   0x25
#define CMD_DOOR_OPEN       0x26
#define CMD_DOOR_CLOSED     0x27
#define CMD_READ_SENSOR     0x28
#define CMD_SENSOR_REPORT   0x29
    
#define CMD_STOP_AUDIO      0x30
#define CMD_PLAY_AUDIO      0x31
#define CMD_ENQ_AUDIO       0x32
#define CMD_LOOP_AUDIO      0x33
#define CMD_SET_AUDIO_VOL   0x34

#define CMD_SERIAL_DATA     0x40

#define ERR_INVALID_CMD         0   // received an unknown command
#define ERR_CMD_TOO_SMALL       1   // got a known command, but not enough data was provided
#define ERR_INVALID_CMD_CONTEXT 2   // got a known command, but it's not valid for the configured device type
#define ERR_AUDIO               3   // audio node couldn't play requested track
#define ERR_INVALID_PARAM       4   // invalid parameters
#define ERR_NO_SERIAL_RESPONSE  5   // sensor didn't respond to a serial request

#define PACKET_SIZE_MIN     6       // smallest NaliG packet size
#define PACKET_SIZE_ACK     6       // size of an ACK packet (no command or data)
#define PACKET_SIZE_CMD     8       // size of a normal NaligPacket (no jumbo data)
#define PACKET_SIZE_MAX     16      // max packet size
#define PACKET_SIZE_HS      23      // 6 byte header + 1 byte command + 16 byte data

#define PROCESS_PACKETS         1
#define DO_NOT_PROCESS_PACKETS  0

#define MAX_SENSOR_CMD_ATTEMPTS 3

// picks correct frequency based on destination
#define PICKFREQ(DEST) ( (DEST == ADDR_MASTER) ? FREQ_B : FREQ_A )

#define PACKET_CACHE_SIZE 10

// NaliG Protocol Standard Packet
// used for all standard messages
// total size is 16 bytes, only 6 are required.  data bytes are optional
struct type_nalig_msg
{
    unsigned char dest_addr;        // destination node address OR destination type (depending on broadcast flag)
    unsigned char source_addr;      // source node address

    unsigned char TTL       : 4;    // time to live (remaining hops) LSB
    unsigned char handshake : 1;    // reserved
    unsigned char relay     : 1;    // allow relaying of this message?
    unsigned char ack       : 1;    // to ack a packet
    unsigned char broadcast : 1;    // broadcast flag  MSB

    unsigned int sequence_id;       // packet sequence ID
    unsigned char packet_id;        // packet ID
    unsigned char data[249];        // data bytes
} __attribute__ ((packed));

typedef struct type_nalig_msg nalig_msg;

struct cacheEntry
{
    unsigned char dest_addr;
    unsigned char source_addr;
    unsigned int sequence_id;
    unsigned char packet_id;

    struct cacheEntry *next;
    struct cacheEntry *prev;
    unsigned char valid;
};

void makePacket(nalig_msg *packet, unsigned char dest_addr, unsigned char flags, unsigned char command, unsigned char data);
void increasePacket(nalig_msg *packet);
void increaseSequence(nalig_msg *packet);
unsigned char processPacket(nalig_msg *packet, unsigned char num_bytes);
unsigned char receiveNaligPacket(unsigned char freq, unsigned char doPacketProcess);
unsigned char sendNaligPacket(unsigned char freq, nalig_msg* packet, unsigned char length);

void initCache();
unsigned char findPacketInCache(nalig_msg *packet);
void addPacketToCache(nalig_msg *packet);
void printPacketCache();

#ifdef	__cplusplus
}
#endif

#endif	/* NALIG_PROTOCOL_H */

