/* 
 * File:   rf.h
 * Author: Daniel
 *
 * Created on January 14, 2013, 9:44 AM
 */



#ifndef RF_H
#define	RF_H

#ifdef	__cplusplus
extern "C" {
#endif

#define CHANNEL_ADR			0x00
#define TX_LENGTH_ADR			0x01
#define TX_CTRL_ADR			0x02
#define TX_CFG_ADR			0x03
#define TX_IRQ_STATUS_ADR		0x04
#define RX_CTRL_ADR			0x05
#define RX_CFG_ADR			0x06
#define RX_IRQ_STATUS_ADR		0x07
#define RX_STATUS_ADR			0x08
#define RX_COUNT_ADR			0x09
#define RX_LENGTH_ADR			0x0A
#define PWR_CTRL_ADR			0x0B
#define XTAL_CTRL_ADR			0x0C
#define IO_CFG_ADR			0x0D
#define GPIO_CTRL_ADR			0x0E
#define XACT_CFG_ADR			0x0F
#define FRAMING_CFG_ADR			0x10
#define DATA32_THOLD_ADR		0x11
#define DATA64_THOLD_ADR		0x12
#define RSSI_ADR			0x13
#define EOP_CTRL_ADR			0x14
#define CRC_SEED_LSB_ADR		0x15
#define CRC_SEED_MSB_ADR		0x16
#define TX_CRC_LSB_ADR			0x17
#define TX_CRC_MSB_ADR			0x18
#define RX_CRC_LSB_ADR			0x19
#define RX_CRC_MSB_ADR			0x1A
#define TX_OFFSET_LSB_ADR		0x1B
#define TX_OFFSET_MSB_ADR		0x1C
#define MODE_OVERRIDE_ADR		0x1D
#define RX_OVERRIDE_ADR			0x1E
#define TX_OVERRIDE_ADR			0x1F
#define XTAL_CFG_ADR			0x26
#define CLK_OFFSET_ADR			0x27
#define CLK_EN_ADR                      0x28
#define RX_ABORT_ADR			0x29
#define AUTO_CAL_TIME_ADR		0x32
#define AUTO_CAL_OFFSET_ADR		0x35
#define ANALOG_CTRL_ADR			0x39

#define TX_BUFFER_ADR			0x20
#define RX_BUFFER_ADR			0x21
#define SOP_CODE_ADR			0x22
#define DATA_CODE_ADR			0x23
#define PREAMBLE_ADR			0x24
#define MFG_ID_ADR			0x25
    
#define SOPPN_LENGTH 24     // this is the total length of the SOP/PN.
#define SOP_LENGTH 8        // this is just the length of the SOP
#define PN_LENGTH 16        // this is just the PN length
#define PAYLOAD_LEN_LG 16   // number of bytes in a large data payload
#define PAYLOAD_LEN_SM 8    // number of bytes in a small data payload

#define NUM_TIMES_RETRANSMIT 5  // number of times to try and re-transmit before giving up

#define RF_TX_FAIL -1
#define RF_IRQ_FAIL -2
#define RF_MIN_PWR  4
#define RF_MAX_PWR  7

#define RSSI_THRESH  5     // RSSI range: 0-31

#define FREQ_A  0
#define FREQ_B  3
#define FREQ_C  6

#define USE_AUTOACK 0x80
#define NO_AUTOACK 0x00

#define readReg(REG)        spiSendCommand(REG, 0x00)
#define writeReg(REG, VAL)  spiSendCommand(REG | 0x80, VAL)

unsigned char CYRF6936_mfgid();
void CYRF6936_init();

void CYRF6936_sleep();
void CYRF6936_wakeup();
void waitForSynth();
//void TXinit();
//void RXinit();
void loadSOPPN();
int TX (unsigned char channel, unsigned char *payload, unsigned char payloadSize);
int RX (unsigned char channel, unsigned char *payload, unsigned int timeout);
unsigned char sleepUntilIRQ(unsigned int timeout);
int TXack();

#ifdef	__cplusplus
}
#endif

#endif	/* RF_H */

