#include <xc.h>
#include "main.h"
#include "spi.h"

unsigned char spiSendByte(unsigned char data)
{
    // send and receive a single byte
    unsigned char result;

    SPI_CS = 0;     // set CS low
    result = spiTransfer(data); // send byte
    SPI_CS = 1;     // set CS high

    return result; 
}

unsigned char spiSendCommand(unsigned char command, unsigned char data)
{
    // send a single command byte, then a single data byte.  return data from second byte
    unsigned char result;

    SPI_CS = 0;     // set CS low
    spiTransfer(command);       // send command byte
    result = spiTransfer(data); // send data byte

    SPI_CS = 1;     // set CS high

    return result;
}

unsigned char spiSendCommand2(unsigned char command, unsigned char data, unsigned char data2)
{
    // send a single command byte, then two data bytes.  return data from third byte
    unsigned char result;

    SPI_CS = 0;     // set CS low
    spiTransfer(command);           // send command byte
    spiTransfer(data);              // send first data byte
    result = spiTransfer(data2);    // send second data byte

    SPI_CS = 1;     // set CS high

    return result;
}

unsigned char spiSendBurstCommand(unsigned char command, unsigned char* data, unsigned char length)
{
    // send multiple bytes after sending a single command byte
    unsigned char result;
    unsigned char i;

    SPI_CS = 0;     // set CS low
    spiTransfer(command);           // send command byte

    // send data bytes
    for (i=0; i<length; i++) {
        result = spiTransfer(data[i]);
    }

    SPI_CS = 1;     // set CS high

    return result;  
}

unsigned char spiReceiveBurstData(unsigned char command, unsigned char* data, unsigned char length)
{
    // receive multiple bytes after sending a single command byte
    unsigned char result;
    unsigned char i;
    SPI_CS = 0;     // set CS low

    spiTransfer(command);           // send command byte

    // send data bytes
    for (i=0; i<length; i++) {
        result = spiTransfer(data[i]);

        data[i] = result;
    }

    SPI_CS = 1;     // set CS high

    return result;
}

unsigned char spiTransfer(volatile unsigned char data)
{
    // send and receieve one byte over SPI (does not control CS line)
    
    SPI1BUF = data; // start transmission

    while (!SPI1STATbits.SPIRBF) { };    // wait for end of transmission

    return SPI1BUF; // return received byte
}