#include <xc.h>
#include <stdio.h>

#include "main.h"
#include "rf.h"
#include "nalig_protocol.h" 
#include "sys_functions.h"

const char HEXITS[16] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};

unsigned char rnd_base[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
unsigned char rnd_counter;

extern unsigned char uartFlag;
extern unsigned char mfg_id[6];

volatile unsigned char tx_buffer[TX_BUFFER_SIZE];
volatile unsigned char tx_buffer_pos=0, tx_buffer_len=0, tx_fifo=0;

extern volatile unsigned char UART_RX_buffer[UART_BUFFER_SIZE];
extern volatile unsigned char UART_RX_buffer_size;

extern sys_config myConfig;
extern unsigned char deviceIsKeyfob;
extern unsigned char unsavedConfigChanges;
extern volatile unsigned char timer2_flag;

unsigned int readBatteryVoltage()
{
    unsigned int ADCresult;
    
    AD1CON1bits.ADON = 1;   // enable ADC
    AD1CON1bits.SAMP = 1;   // enable ADC
    IFS0bits.AD1IF = 0;
    //while (!AD1CON1bits.DONE); // || AD1CON1bits.SAMP) { }
    while(!IFS0bits.AD1IF);  //waiting
    //IFS0bits.AD1IF = 0;
    ADCresult = ADC1BUF0;

    AD1CON1bits.ADON = 0;

    return ADCresult;
}

unsigned char compareArrays(unsigned char *array1, unsigned char *array2, int length)
{
  // compare 2 unsigned char arrays.  if they are exactly the same, return 1.  else, 0
  unsigned char i;

  for (i=0; i<length; i++)
  {
    if (array1[i] != array2[i])
      return 0;
  }
  return 1;
}


void txChar(unsigned char c)
{
    if (deviceIsKeyfob || myConfig.myType == TYPE_SERIAL) return;    // keyfobs can't serial
    // Serial bridge devices should not send ANY other data, so we are effectively disabling serial other than serial data

    //while ( U1STAbits.UTXBF || U1STAbits.TRMT == 0);   // wait while Tx buffer full
    //U1TXREG = c;

        // normally we would use UTXBF to determine if the buffer is full, but there is a silicon error on the 24F16KA301 family
    // so we have to manually keep track of the TX FIFO (should be able to put 4 bytes in, but we can only put in 3)
    /*
    if (tx_fifo >= 2 || (tx_buffer_len > 0))// transmit buffer is full or we already have buffer characters
    {
        while (tx_buffer_len > TX_BUFFER_SIZE) ; // if we fill up our TX buffer, well that sucks.
            tx_buffer[tx_buffer_len++] = c;
    } else {
        U1TXREG = c;
        tx_fifo++;
    }*/

    if (U1STAbits.TRMT == 1 ) // can we just transmit this character?
    {
        U1TXREG = c;
    } else {     // hardware buffer is full or we already have data in our software buffer
        while (tx_buffer_len >= TX_BUFFER_SIZE) ;   // if our software TX buffer is full, we will wait until it is empty.
        tx_buffer[tx_buffer_len++] = c; // add this character to our buffer
    }

}

void txStr(char *s)
{
    while( *s)			// loop until *s == '\0' the  end of the string
        txChar( *s++);	// send the character and point to the next one
}

void txHex(unsigned char in)
{
    // send a char as two hexits
    txChar( HEXITS[((0xF0 & in)>>4)] );
    txChar( HEXITS[(0x0F & in)] );
    //txChar( '\r');       // terminate with a cr / line feed
}

void txInt(unsigned int in)
{
    // send a char as two hexits
    txChar( HEXITS[((0xF000 & in)>>12)] );
    txChar( HEXITS[((0x0F00 & in)>>8)] );
    txChar( HEXITS[((0x00F0 & in)>>4)] );
    txChar( HEXITS[ (0x000F & in)] );
    //txChar('\t');
    //txChar( '\r');       // terminate with a cr / line feed
}

void txVal(char *s, unsigned char val)
{
    txStr(s);
    txHex(val);
    txChar('\r');
}

void txVal2(char *s, unsigned char val, unsigned char val2)
{
    txStr(s);
    txHex(val);
    txStr(",");
    txHex(val2);
    txChar('\r');
}

void txReport(unsigned char source, char *s, unsigned char val)
{
    txStr(s);
    txHex(source);
    txHex(val);
    txChar('\r');
}

void txReport2(unsigned char source, char *s, unsigned char val, unsigned char val2)
{
    txStr(s);
    txHex(source);
    txHex(val);
    txStr(",");
    txHex(val2);
    txChar('\r');
}

void txArray(unsigned char *array, unsigned char length)
{
    unsigned char i;
    for (i=0; i < length; i++)
    {
        txHex(array[i]);
        txChar(' ');
    }
}

void clearArray(unsigned char *array, unsigned char length)
{
    unsigned char i;
    for (i=0; i < length; i++)
    {
        array[i] = 0;
    }
}

unsigned char inArray(unsigned char search, unsigned char *array, unsigned char length)
{
    unsigned char i;
    for (i=0; i < length; i++)
    {
        if (array[i] == search) return 1;
    }
    return 0;
}

void copyArray(unsigned char *dest, unsigned char *src, unsigned char length)
{
    unsigned char i;
    for (i=0; i < length; i++)
    {
        dest[i] = src[i];
    }
}

void generateNonce(unsigned char *output)
{
    // generates a 16 byte pseudo random number
    unsigned char i, z, y;
    z = TMR1 & 0x7;
    y = 1;
    for (i=0; i<16; i++)
    {
        if (z > 5) z = 0;
        output[i] = mfg_id[z] + (TMR1 & y) - (TMR2 & y) + rnd_base[i];
        z++;
        y = y << 1;
    }

}



void rest(unsigned int timeout)
{
    // goes to sleep for some time

    TMR2 = timeout;
    timer2_flag = 0;
    T2CONbits.TON = 1;          // enable timer 2 for a timeout
   
    while (!timer2_flag) {
        Idle();    // go to sleep, wait for timer interrupts ONLY
    }
    Nop();

    T2CONbits.TON = 0;          // disable timer 2
}

unsigned char waitForSerialAck(unsigned int timeout)
{
    // wait for the specified time to receive an "ack" from a sensor, which is just "k\r" (two characters)
    // this is better than trying to do the processing in an interrupt!
    // returns 1 if we got the ack, 0 if not
    
    TMR2 = timeout;
    timer2_flag = 0;
    T2CONbits.TON = 1;          // enable timer 2 for a timeout
    
    while (!timer2_flag)    //any interrupt (including UART RX) will cause this code to execute
    {
        if (UART_RX_buffer_size > 1)    // we gotta have received at least 2 bytes
        {
            if ((UART_RX_buffer[UART_RX_buffer_size-1] == '\r') && (UART_RX_buffer[UART_RX_buffer_size-2] == 'k')) {    // check for k\r
                UART_RX_buffer_size -= 2;   // we got the ack.  no need to process this later
                return 1;   // success
            }
        }
        Idle();    // go to sleep, waiting for an interrupt
    }

    T2CONbits.TON = 0;          // disable timer 2

    return 0; // failure
}


void setGPIO(unsigned char byte1, unsigned char byte2)
{
    /*
    BYTE 1
    Bit 0 - GPIO 0 high (1) / low (0)
    Bit 1 - GPIO 1 high (1) / low (0)
    Bit 2 - GPIO 2 high (1) / low (0)
    Bit 3 - GPIO 3 high (1) / low (0)
    Bit 4 - GPIO 0 In (1) /Out (0)
    Bit 5 - GPIO 1 In (1) /Out (0)
    Bit 6 - GPIO 2 In (1) /Out (0)
    Bit 7 - GPIO 3 In (1) /Out (0)

    BYTE 2
    Bit 0 - GPIO 0 alert enabled
    Bit 1 - GPIO 1 alert enabled
    Bit 2 - GPIO 2 alert enabled
    Bit 3 - GPIO 3 alert enabled
    Bit 4 - GPIO 0 pull-up enabled
    Bit 5 - GPIO 1 pull-up enabled
    Bit 6 - GPIO 2 pull-up enabled
    Bit 7 - GPIO 3 pull-up enabled
    */
    GPIO0_TRIS = (byte1 & 0b00010000) ? 1 : 0;
    GPIO1_TRIS = (byte1 & 0b00100000) ? 1 : 0;
    GPIO2_TRIS = (byte1 & 0b01000000) ? 1 : 0;
    GPIO3_TRIS = (byte1 & 0b10000000) ? 1 : 0;

    GPIO0_PU = (byte2 & 0b00010000) ? 1 : 0;
    GPIO1_PU = (byte2 & 0b00100000) ? 1 : 0;
    GPIO2_PU = (byte2 & 0b01000000) ? 1 : 0;
    GPIO3_PU = (byte2 & 0b10000000) ? 1 : 0;

    GPIO0_LAT = (byte1 & 0b00000001) ? 1 : 0;
    GPIO1_LAT = (byte1 & 0b00000010) ? 1 : 0;
    GPIO2_LAT = (byte1 & 0b00000100) ? 1 : 0;
    GPIO3_LAT = (byte1 & 0b00001000) ? 1 : 0;
    myConfig.GPIO_config1 = byte1;
    myConfig.GPIO_config2 = byte2;
}

void loadDeafultConfig()
{
    myConfig.myAddr = 0xFE;
    myConfig.myType = 0xFE;
    myConfig.addr_checksum = myConfig.myAddr + myConfig.myType;

    myConfig.LED_blink = 1;
    myConfig.relayEnabled = 1;
    myConfig.hasRFE = 0;
    myConfig.unused = 0;
    
    // GPIO default: all pins input, pullup on GPIO0
    myConfig.GPIO_config1 = 0xF0;
    myConfig.GPIO_config2 = 0x10;

    myConfig.lastPacketID = 0;
    myConfig.lastSeqID = 0;
}

unsigned char readDeviceConfig()
{
    // returns 1 if ok, 0 if invalid config detected
    unsigned char i,j=0;
    unsigned int *config;

    config = (unsigned int *)&myConfig;

    for (i = 0; i < 0x0E; i+=2)
    {
        config[j++] = eeRead(i);
    }

    if (myConfig.addr_checksum != (unsigned char)(myConfig.myAddr + myConfig.myType - 1))
    {
        return 0;   //checksum doesn't match
    }

    if (myConfig.myType == TYPE_KEYFOB) deviceIsKeyfob = 1;

    return 1;   // config is valid
}

void saveDeviceConfig()
{
    unsigned char i,j=0;
    unsigned int *config;

    myConfig.addr_checksum = myConfig.myAddr + myConfig.myType-1;
    config = (unsigned int *)&myConfig;

    for (i = 0; i < 0x0E; i+=2)
    {
        eeWrite(i, config[j++]);
    }
}

unsigned int eeRead(unsigned int address)
{
    //while(NVMCONbits.WR);      // wait for any last write to complete
    TBLPAG = 0x7f;            // eprom is a t 0x7ffe00
    return __builtin_tblrdl(0xFE00+address);
}

void eeWrite(unsigned int address, unsigned int data)
{
    while(NVMCONbits.WR);      // wait for last write to complete
    NVMCON = 0x4058;        // Set up NVMCON to erase one word of data EEPROM
    TBLPAG = 0x7f;            // eprom is a t 0x7ffe00
    __builtin_tblwtl(0xFE00+address, 0xffff);
    asm volatile ("disi #5");
    __builtin_write_NVM();       // Issue Unlock Sequence & Start Write Cycle

    while(NVMCONbits.WR);      // wait for errase to complete
    NVMCON = 0x4004;        // Set up NVMCON to write one word of data EEPROM
    __builtin_tblwtl(0xFE00+address, data);   // Write EEPROM data to write latch
    asm volatile ("disi #5");
    __builtin_write_NVM();
    while(NVMCONbits.WR);      // wait for errase to complete
} 