/* 
 * File:   sys_functions.h
 * Author: Daniel
 *
 * Created on March 18, 2013, 8:13 PM
 */

#ifndef SYS_FUNCTIONS_H
#define	SYS_FUNCTIONS_H

#ifdef	__cplusplus
extern "C" {
#endif

// timeouts for timer 2 (with Fosc = 4 Mhz, 1:1 prescale)
#define TIMEOUT_32MS    0
#define TIMEOUT_20MS    25535
#define TIMEOUT_10MS    45535       // 65535 - (10e-3*2e6)
#define TIMEOUT_5MS     55535
#define TIMEOUT_2MS     61535
#define TIMEOUT_1MS     63535
#define TIMEOUT_500US   64535
#define TIMEOUT_200US   65135
#define TIMEOUT_100US   65335

// EEPROM addresses
#define EE_MYADDR_ADDR      0x00
#define EE_CHKSUM_ADDR      0x02
#define EE_MYTYPE_ADDR      0x10
#define EE_SETTINGS_ADDR    0x04
#define EE_GPIO_ADDR        0x06

#define EE_CONFIG_START     0x00        // start of config
#define EE_CONFIG_LEN       0x0F        // length of config



struct type_system_config   // length of this struct must match EE_CONFIG_LEN
{
    unsigned char myAddr;               // device address
    unsigned char myType;               // device type
    unsigned char addr_checksum;        // checksum: device address + device type (for verifying integrity)

    unsigned char LED_blink     : 1;    // LED blink-alive
    unsigned char relayEnabled  : 1;    // packet relay
    unsigned char hasRFE        : 1;    // device has CC2591 RF Front End
    unsigned char unused        : 5;    // reserved bits for future use

    unsigned char GPIO_config1;         // GPIO TRIS and LAT data
    unsigned char GPIO_config2;         // GPIO pullup and alert data

    unsigned char lastPacketID;         // start packet ID at this number on startup
    unsigned int lastSeqID;             // start squence ID at this number on startup

    unsigned char more_unused[7];       // reserved bytes for future use
    
} __attribute__ ((packed));

typedef struct type_system_config sys_config;


#define SMALL_RND_DELAY rest(65535 - (TMR4 & 0xFF))
#define addRndBase(x) rnd_base[(rnd_counter++ & 0xF)] = x
    
unsigned int readBatteryVoltage();
void generateNonce(unsigned char *output);
void rest(unsigned int timeout);
unsigned char waitForSerialAck(unsigned int timeout);
void setGPIO(unsigned char byte1, unsigned char byte2);

void txChar(unsigned char c);
void txStr( char *s);
void txHex(unsigned char in);
void txInt(unsigned int in);
void txVal(char *s, unsigned char val);
void txVal2(char *s, unsigned char val, unsigned char val2);
void txReport(unsigned char source, char *s, unsigned char val);
void txReport2(unsigned char source, char *s, unsigned char val, unsigned char val2);
void txArray(unsigned char *array, unsigned char length);

void clearArray(unsigned char *array, unsigned char length);
unsigned char inArray(unsigned char search, unsigned char *array, unsigned char length);
void copyArray(unsigned char *dest, unsigned char *src, unsigned char length);
unsigned char compareArrays(unsigned char *array1, unsigned char *array2, int length);

unsigned char readDeviceConfig();
void saveDeviceConfig();
void loadDeafultConfig();

void eeWrite(unsigned int address, unsigned int data);
unsigned int eeRead(unsigned int address);
#ifdef	__cplusplus
}
#endif

#endif	/* SYS_FUNCTIONS_H */

